import os
import logging

from enum import IntEnum
from appdirs import user_data_dir

DATA_DIR = user_data_dir('xbox', 'OpenXbox')
TOKENS_FILE = os.path.join(DATA_DIR, 'tokens.json')
CONSOLES_FILE = os.path.join(DATA_DIR, 'consoles.json')

if not os.path.exists(DATA_DIR):
    os.makedirs(DATA_DIR)

LOG_FMT = '[%(asctime)s] %(levelname)s: %(message)s'
LOG_LEVEL_DEBUG_INCL_PACKETS = logging.DEBUG - 1
logging.addLevelName(LOG_LEVEL_DEBUG_INCL_PACKETS, 'DEBUG_INCL_PACKETS')


class ExitCodes(IntEnum):
    """
    Common CLI exit codes
    """
    OK = 0
    ArgParsingError = 1
    AuthenticationError = 2
    DiscoveryError = 3
    ConsoleChoice = 4


class VerboseFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        super(VerboseFormatter, self).__init__(*args, **kwargs)
        self._verbosefmt = self._fmt + '\n%(_msg)s'

    def formatMessage(self, record):
        if '_msg' in record.__dict__:
            return self._verbosefmt % record.__dict__
        return self._style.format(record)



from flask import current_app as app
from . import routes


@routes.route('/')
def index():
    routes = set([str(rule) for rule in app.url_map.iter_rules()])
    return app.success(endpoints=sorted(routes))



from xbox.sg.crypto import PKCS7Padding
from xbox.auxiliary.packet import aux_header_struct, AUX_PACKET_MAGIC


class AuxiliaryPackerException(Exception):
    pass


def pack(data, crypto, server_data=False):
    """
    Encrypt auxiliary data blob

    Args:
        data (bytes): Data
        crypto (:class:`AuxiliaryStreamCrypto`): Crypto context
        server_data (bool): Whether to encrypt with `server IV`

    Returns:
        bytes: Encrypted message
    """
    # Store payload size without padding
    payload_size = len(data)

    # Pad data
    padded = PKCS7Padding.pad(data, 16)

    if not server_data:
        ciphertext = crypto.encrypt(padded)
    else:
        ciphertext = crypto.encrypt_server(padded)

    header = aux_header_struct.build(dict(
        magic=AUX_PACKET_MAGIC,
        payload_size=payload_size)
    )

    msg = header + ciphertext
    hmac = crypto.hash(msg)
    msg += hmac

    messages = list()
    while len(msg) > 1448:
        fragment, msg = msg[:1448], msg[1448:]
        messages.append(fragment)

    messages.append(msg)

    return messages


def unpack(data, crypto, client_data=False):
    """
    Split and decrypt auxiliary data blob

    Args:
        data (bytes): Data blob
        crypto (:class:`AuxiliaryStreamCrypto`): Crypto context
        client_data (bool): Whether to decrypt with 'client IV'

    Returns:
        bytes: Decrypted message
    """
    # Split header from rest of data
    header, payload, hmac = data[:4], data[4:-32], data[-32:]

    parsed = aux_header_struct.parse(header)

    if not crypto.verify(header + payload, hmac):
        raise AuxiliaryPackerException('Hash verification failed')

    if not client_data:
        plaintext = crypto.decrypt(payload)
    else:
        plaintext = crypto.decrypt_client(payload)

    # Cut off padding, before returning
    return plaintext[:parsed.payload_size]



"""
Smartglass enumerations
"""

from enum import Enum, Flag


class PacketType(Enum):
    """
    Packet Types of core protocol:
    Connect*, Discovery*, PowerOnRequest -> SimpleMessage
    Message -> Message
    """
    ConnectRequest = 0xCC00
    ConnectResponse = 0xCC01
    DiscoveryRequest = 0xDD00
    DiscoveryResponse = 0xDD01
    PowerOnRequest = 0xDD02
    Message = 0xD00D


class ClientType(Enum):
    """
    Client Type

    Used in `DiscoveryRequest`, `DiscoveryResponse` and `LocalJoin`-Message.
    """
    XboxOne = 0x1
    Xbox360 = 0x2
    WindowsDesktop = 0x3
    WindowsStore = 0x4
    WindowsPhone = 0x5
    iPhone = 0x6
    iPad = 0x7
    Android = 0x8


class DeviceCapabilities(Flag):
    """
    Bitmask for client device hardware capabilities
    """
    Non = 0
    Streaming = 1
    Audio = 2
    Accelerometer = 4
    Compass = 8
    Gyrometer = 16
    Inclinometer = 32
    Orientation = 64
    All = 0xFFFFFFFFFFFFFFFF


class ConnectionResult(Enum):
    """
    Connection Result

    Used in result-field of `ConnectResponse` packet.
    """
    Success = 0x0
    Pending = 0x1
    FailureUnknown = 0x2
    FailureAnonymousConnectionsDisabled = 0x3
    FailureDeviceLimitExceeded = 0x4
    FailureSmartGlassDisabled = 0x5
    FailureUserAuthFailed = 0x6
    FailureUserSignInFailed = 0x7
    FailureUserSignInTimeOut = 0x8
    FailureUserSignInRequired = 0x9


class PublicKeyType(Enum):
    """
    Public Key Type

    Used in `ConnectRequest` packet.
    """
    EC_DH_P256 = 0x00
    EC_DH_P384 = 0x01
    EC_DH_P521 = 0x02
    Default = 0xFFFF


class AckStatus(Enum):
    """
    Acknowledgement Status

    Internally used to signalize the waiting
    caller the status of acknowledgement.
    """
    Pending = 0
    Processed = 1
    Rejected = 2


class ConnectionState(Enum):
    """
    Connection State

    Internally used by :obj:`.Console`
    """
    Disconnected = 0x0
    Connecting = 0x1
    Connected = 0x2
    Error = 0x3
    Disconnecting = 0x4
    Reconnecting = 0x5


class DeviceStatus(Enum):
    """
    Device Status

    Internally used by :obj:`.Console`
    """
    DiscoveringAvailability = 0x1
    Available = 0x2
    Unavailable = 0x3


class PrimaryDeviceFlag(Flag):
    """
    Primary Device Flag

    Used in flags of `DiscoveryResponse` packet.
    """
    Null = 0x0
    AllowConsoleUsers = 0x1
    AllowAuthenticatedUsers = 0x2
    AllowAnonymousUsers = 0x4
    CertificatePending = 0x8


class DisconnectReason(Enum):
    """
    Disconnect Reason

    Used in reason-field of `Disconnect`-Message.
    """
    Unspecified = 0x0
    Error = 0x1
    PowerOff = 0x2
    Maintenance = 0x3
    AppClose = 0x4
    SignOut = 0x5
    Reboot = 0x6
    Disabled = 0x7
    LowPower = 0x8


class PairedIdentityState(Enum):
    """
    Paired Identity State

    Used in `ConnectResponse` and `PairedIdentityStateChanged`-Message.
    """
    NotPaired = 0x0
    Paired = 0x1


class ServiceChannel(Enum):
    """
    Service Channels

    Used internally to identify actual Channel Ids.
    """
    Core = 0x0
    SystemInput = 0x1
    SystemInputTVRemote = 0x2
    SystemMedia = 0x3
    SystemText = 0x4
    SystemBroadcast = 0x5
    Ack = 0x6
    Title = 0x7


class MessageType(Enum):
    """
    Message Type

    Used in `Message`-Header.
    """
    Null = 0x0
    Ack = 0x1
    Group = 0x2
    LocalJoin = 0x3
    StopActivity = 0x5
    AuxilaryStream = 0x19
    ActiveSurfaceChange = 0x1a
    Navigate = 0x1b
    Json = 0x1c
    Tunnel = 0x1d
    ConsoleStatus = 0x1e
    TitleTextConfiguration = 0x1f
    TitleTextInput = 0x20
    TitleTextSelection = 0x21
    MirroringRequest = 0x22
    TitleLaunch = 0x23
    StartChannelRequest = 0x26
    StartChannelResponse = 0x27
    StopChannel = 0x28
    System = 0x29
    Disconnect = 0x2a
    TitleTouch = 0x2e
    Accelerometer = 0x2f
    Gyrometer = 0x30
    Inclinometer = 0x31
    Compass = 0x32
    Orientation = 0x33
    PairedIdentityStateChanged = 0x36
    Unsnap = 0x37
    GameDvrRecord = 0x38
    PowerOff = 0x39
    MediaControllerRemoved = 0xf00
    MediaCommand = 0xf01
    MediaCommandResult = 0xf02
    MediaState = 0xf03
    Gamepad = 0xf0a
    SystemTextConfiguration = 0xf2b
    SystemTextInput = 0xf2c
    SystemTouch = 0xf2e
    SystemTextAck = 0xf34
    SystemTextDone = 0xf35


class ActiveTitleLocation(Enum):
    """
    Active Title Location
    """
    Full = 0x0
    Fill = 0x1
    Snapped = 0x2
    StartView = 0x3
    SystemUI = 0x4
    Default = 0x5


class ActiveSurfaceType(Enum):
    """
    Active Surface Type

    Used in `ActiveSurfaceType`-Message.
    """
    Blank = 0x0
    Direct = 0x1
    HTML = 0x2
    TitleTextEntry = 0x3


class MediaType(Enum):
    """
    Media Type

    Used in `MediaState`-Message.
    """
    NoMedia = 0x0
    Music = 0x1
    Video = 0x2
    Image = 0x3
    Conversation = 0x4
    Game = 0x5


class MediaTransportState(Enum):
    """
    Media Transport State

    Used in `MediaState`-Message.
    """
    Invalid = 0x0
    Stopped = 0x1
    Starting = 0x2
    Playing = 0x3
    Paused = 0x4
    Buffering = 0x5


class MediaPlaybackStatus(Enum):
    """
    Media Playback Status

    Used in `MediaState`-Message.
    """
    Closed = 0x0
    Changing = 0x1
    Stopped = 0x2
    Playing = 0x3
    Paused = 0x4


class MediaControlCommand(Flag):
    """
    Media Control Command

    Used in `MediaCommand`-Message.
    """
    Null = 0x0
    Play = 0x2
    Pause = 0x4
    PlayPauseToggle = 0x8
    Stop = 0x10
    Record = 0x20
    NextTrack = 0x40
    PreviousTrack = 0x80
    FastForward = 0x100
    Rewind = 0x200
    ChannelUp = 0x400
    ChannelDown = 0x800
    Back = 0x1000
    View = 0x2000
    Menu = 0x4000
    Seek = 0x8000


class SoundLevel(Enum):
    """
    Sound Level

    Used in `MediaState`-Message.
    """
    Muted = 0x0
    Low = 0x1
    Full = 0x2


class GamePadButton(Enum):
    """
    Gamepad Button

    Used in `Gamepad`-Message.
    """
    Clear = 0x0
    Enroll = 0x1
    Nexu = 0x2
    Menu = 0x4
    View = 0x8
    PadA = 0x10
    PadB = 0x20
    PadX = 0x40
    PadY = 0x80
    DPadUp = 0x100
    DPadDown = 0x200
    DPadLeft = 0x400
    DPadRight = 0x800
    LeftShoulder = 0x1000
    RightShoulder = 0x2000
    LeftThumbStick = 0x4000
    RightThumbStick = 0x8000


class TouchAction(Enum):
    """
    Touch Action

    Used in `SystemTouch`/`TitleTouch`-Message.
    """
    Null = 0x0
    Down = 0x1
    Move = 0x2
    Up = 0x3
    Cancel = 0x4


class TextInputScope(Enum):
    """
    Text Input Scope

    Used in `TextConfiguration`-Message
    """
    Default = 0x0
    Url = 0x1
    FullFilePath = 0x2
    FileName = 0x3
    EmailUserName = 0x4
    EmailSmtpAddress = 0x5
    LogOnName = 0x6
    PersonalFullName = 0x7
    PersonalNamePrefix = 0x8
    PersonalGivenName = 0x9
    PersonalMiddleName = 0xa
    PersonalSurname = 0xb
    PersonalNameSuffix = 0xc
    PostalAddress = 0xd
    PostalCode = 0xe
    AddressStreet = 0xf
    AddressStateOrProvince = 0x10
    AddressCity = 0x11
    AddressCountryName = 0x12
    AddressCountryShortName = 0x13
    CurrencyAmountAndSymbol = 0x14
    CurrencyAmount = 0x15
    Date = 0x16
    DateMonth = 0x17
    DateDay = 0x18
    DateYear = 0x19
    DateMonthName = 0x1a
    DateDayName = 0x1b
    Digits = 0x1c
    Number = 0x1d
    OneChar = 0x1e
    Password = 0x1f
    TelephoneNumber = 0x20
    TelephoneCountryCode = 0x21
    TelephoneAreaCode = 0x22
    TelephoneLocalNumber = 0x23
    Time = 0x24
    TimeHour = 0x25
    TimeMinorSec = 0x26
    NumberFullWidth = 0x27
    AlphanumericHalfWidth = 0x28
    AlphanumericFullWidth = 0x29
    CurrencyChinese = 0x2a
    Bopomofo = 0x2b
    Hiragana = 0x2c
    KatakanaHalfWidth = 0x2d
    KatakanaFullWidth = 0x2e
    Hanja = 0x2f
    HangulHalfWidth = 0x30
    HangulFullWidth = 0x31
    Search = 0x32
    SearchTitleText = 0x33
    SearchIncremental = 0x34
    ChineseHalfWidth = 0x35
    ChineseFullWidth = 0x36
    NativeScript = 0x37
    Unknown = 0x39


class TextAction(Enum):
    """
    Text Action
    """
    Cancel = 0x0
    Accept = 0x1


class TextOption(Flag):
    """
    Text Option
    """
    Default = 0x0
    AcceptsReturn = 0x1
    Password = 0x2
    MultiLine = 0x4
    SpellCheckEnabled = 0x8
    PredictionEnabled = 0x10
    RTL = 0x20
    Dismiss = 0x4000


class TextResult(Enum):
    """
    Text Result
    """
    Cancel = 0x0
    Accept = 0x1
    Null = 0xFFFF


class HashAlgorithm(Enum):
    """
    Hash Algorithm

    Unused
    """
    SHA256 = 0x0
    SHA384 = 0x1
    SHA512 = 0x2


class AsymmetricAlgorithm(Enum):
    """
    Asymmetric Algorithm

    Unused
    """
    RSA_PKCS1_1024 = 0x0
    RSA_OAEP_1024 = 0x1
    RSA_PKCS1_2048 = 0x2
    RSA_OAEP_2048 = 0x3
    EC_DSA_P256 = 0x4
    EC_DSA_P384 = 0x5
    EC_DSA_P521 = 0x6
    EC_DH_P256 = 0x7
    EC_DH_P384 = 0x8
    EC_DH_P521 = 0x9


class SymmetricAlgorithm(Enum):
    """
    Symmetric Algorithm

    Unused
    """
    AES_CBC_128 = 0x0
    AES_CBC_192 = 0x1
    AES_CBC_256 = 0x2


class AuthError(Enum):
    """
    Authentication Error

    Unused
    """
    Null = 0x0
    DevModeNotAuthorized = 0x2
    SystemUpdateRequired = 0x3
    ContentUpdateRequired = 0x4
    EnforcementBan = 0x5
    ThirdPartyBan = 0x6
    ParentalControlsBan = 0x7
    SubscriptionNotActivated = 0x8
    BillingMaintenanceRequired = 0x9
    AccountNotCreated = 0xa
    NewTermsOfUse = 0xb
    CountryNotAuthorized = 0xc
    AgeVerificationRequired = 0xd
    Curfew = 0xe
    ChildAccountNotInFamily = 0xf
    CSVTransitionRequired = 0x10
    AccountMaintenanceRequired = 0x11
    AccountTypeNotAllowed = 0x12
    ContentIsolation = 0x13
    GamertagMustChange = 0x14
    DeviceChallengeRequired = 0x15
    SignInCountExceeded = 0x16
    RetailAccountNotAllowed = 0x17
    SandboxNotAllowed = 0x18
    UnknownUser = 0x19
    RetailContentNotAuthorized = 0x1a
    ContentNotAuthorized = 0x1b
    ExpiredDeviceToken = 0x1c
    ExpiredTitleToken = 0x1d
    ExpiredUserToken = 0x1e
    InvalidDeviceToken = 0x1f
    InvalidTitleToken = 0x20
    InvalidUserToken = 0x21
    InvalidRefreshToken = 0x22


class TokensResetReason(Enum):
    """
    Tokens Reset Reason

    Unused
    """
    AuthTicketError = 0x0
    AuthTicketChanged = 0x1
    EnvironmentChanged = 0x2
    SandboxIDChanged = 0x3


class HttpRequestMethod(Enum):
    """
    HTTP Request Method

    Unused
    """
    GET = 0x0
    POST = 0x1


class EnvironmentType(Enum):
    """
    Environment Type

    Unused
    """
    Production = 0x0
    DNet = 0x1
    Mock = 0x2
    Null = 0x3


class MetricsOrigin(Enum):
    """
    Metrics Origin

    Unused
    """
    Core = 0x1
    SDK = 0x2
    Canvas = 0x3
    App = 0x4


class SGResultCode(Enum):
    """
    Smartglass Result Code

    Unused
    """
    SG_E_SUCCESS = 0x0
    SG_E_ABORT = 0x80000004
    SG_E_ACCESS_DENIED = 0x80000005
    SG_E_FAIL = 0x80000006
    SG_E_HANDLE = 0x80000007
    SG_E_INVALID_ARG = 0x80000008
    SG_E_NO_INTERFACE = 0x80000009
    SG_E_NOT_IMPL = 0x8000000a
    SG_E_OUT_OF_MEMORY = 0x8000000b
    SG_E_POINTER = 0x8000000c
    SG_E_UNEXPECTED = 0x8000000d
    SG_E_PENDING = 0x8000000e
    SG_E_INVALID_DATA = 0x8000000f
    SG_E_CANCELED = 0x80000010
    SG_E_INVALID_STATE = 0x80000011
    SG_E_NOT_FOUND = 0x80000012
    SG_E_NO_MORE_CAPACITY = 0x80000013
    SG_E_FAILED_TO_START_THREAD = 0x80000014
    SG_E_MESSAGE_EXPIRED = 0x80000015
    SG_E_TIMED_OUT = 0x80000016
    SG_E_NOT_INITIALIZED = 0x80000017
    SG_E_JSON_LENGTH_EXCEEDED = 0x80000018
    SG_E_MESSAGE_LENGTH_EXCEEDED = 0x80000019
    SG_E_INVALID_CONFIGURATION = 0x8000001a
    SG_E_EXPIRED_CONFIGURATION = 0x8000001b
    SG_E_AUTH_REQUIRED = 0x8000001d
    SG_E_TIMED_OUT_PRESENCE = 0x8000001e
    SG_E_TIMED_OUT_CONNECT = 0x8000001f
    SG_E_SOCKET_ERROR = 0x80010001
    SG_E_HTTP_ERROR = 0x80020001
    SG_E_CANCEL_SHUTDOWN = 0x80020002
    SG_E_HTTP_STATUS = 0x80020003
    SG_E_UNEXPECTED_CRYPTO_ERROR = 0x80030001
    SG_E_INVALID_CRYPT_ARG = 0x80030002
    SG_E_CRYPTO_INVALID_SIGNATURE = 0x80030003
    SG_E_INVALID_CERTIFICATE = 0x80030004
    SG_E_CHANNEL_REQUEST_UNKNOWN_ERROR = 0x80040105
    SG_E_FAILED_TO_JOIN = 0x80060001
    SG_E_ALREADY_CONNECTED = 0x80060002
    SG_E_NOT_CONNECTED = 0x80060003
    SG_E_CONSOLE_NOT_RESPONDING = 0x80060004
    SG_E_CONSOLE_DISCONNECTING = 0x80060005
    SG_E_BIG_ENDIAN_STREAM_STRING_NOT_TERMINATED = 0x80070001
    SG_E_CHANNEL_ALREADY_STARTED = 0x80080001
    SG_E_CHANNEL_FAILED_TO_START = 0x80080002
    SG_E_MAXIMUM_CHANNELS_STARTED = 0x80080003
    SG_E_JNI_CLASS_NOT_FOUND = 0x80090001
    SG_E_JNI_METHOD_NOT_FOUND = 0x80090002
    SG_E_JNI_RUNTIME_ERROR = 0x80090003



import pytest

from xbox.sg.enum import ServiceChannel
from xbox.auxiliary.manager import TitleManager, TitleManagerError


def test_manager_messagehandling(console, decrypted_packets):
    manager = TitleManager(console)

    def handle_msg(msg):
        manager._pre_on_message(msg, ServiceChannel.Title)

    assert manager.active_surface is None
    assert manager.connection_info is None

    # Send unpacked msgs to manager
    handle_msg(decrypted_packets['active_surface_change'])
    handle_msg(decrypted_packets['auxiliary_stream_connection_info'])

    invalid_msg = decrypted_packets['auxiliary_stream_hello']
    invalid_msg.header.flags(msg_type=0x3)
    with pytest.raises(TitleManagerError):
        handle_msg(invalid_msg)

    assert manager.active_surface == decrypted_packets['active_surface_change'].protected_payload
    assert manager.connection_info == decrypted_packets['auxiliary_stream_connection_info'].protected_payload.connection_info



import pytest


@pytest.mark.script_launch_mode('subprocess')
def test_cli_rest_server(script_runner):
    """
    Needs to be done in subprocess due to monkey-patching
    """
    ret = script_runner.run('xbox-rest-server', '--help')
    assert ret.success


def test_cli_maincli(script_runner):
    ret = script_runner.run('xbox-cli', '--help')
    assert ret.success


def test_cli_discover(script_runner):
    ret = script_runner.run('xbox-discover', '--help')
    assert ret.success


def test_cli_poweron(script_runner):
    ret = script_runner.run('xbox-poweron', '--help')
    assert ret.success


def test_cli_poweroff(script_runner):
    ret = script_runner.run('xbox-poweroff', '--help')
    assert ret.success


def test_cli_repl(script_runner):
    ret = script_runner.run('xbox-repl', '--help')
    assert ret.success


def test_cli_replserver(script_runner):
    ret = script_runner.run('xbox-replserver', '--help')
    assert ret.success


def test_cli_textinput(script_runner):
    ret = script_runner.run('xbox-textinput', '--help')
    assert ret.success


def test_cli_gamepadinput(script_runner):
    ret = script_runner.run('xbox-gamepadinput', '--help')
    assert ret.success


def test_cli_tui(script_runner):
    ret = script_runner.run('xbox-tui', '--help')
    assert ret.success


def test_cli_falloutrelay(script_runner):
    ret = script_runner.run('xbox-fo4-relay', '--help')
    assert ret.success


def test_cli_pcap(script_runner):
    ret = script_runner.run('xbox-pcap', '--help')
    assert ret.success


def test_cli_recrypt(script_runner):
    ret = script_runner.run('xbox-recrypt', '--help')
    assert ret.success



import logging

from xbox.sg import factory
from xbox.sg.manager import Manager
from xbox.sg.enum import ServiceChannel, MessageType
from xbox.sg.constants import MessageTarget
from xbox.sg.utils.events import Event

log = logging.getLogger(__name__)


class TitleManagerError(Exception):
    """
    Exception thrown by TitleManager
    """
    pass


class TitleManager(Manager):
    __namespace__ = 'title'

    def __init__(self, console):
        """
        Title Manager (ServiceChannel.Title)

        Args:
             console (:class:`.Console`): Console object, internally passed
                                          by `Console.add_manager`
        """
        super(TitleManager, self).__init__(console, ServiceChannel.Title)
        self._active_surface = None
        self._connection_info = None

        self.on_surface_change = Event()
        self.on_connection_info = Event()

    def _on_message(self, msg, channel):
        """
        Internal handler method to receive messages from Title Channel

        Args:
            msg (:class:`XStructObj`): Message
            channel (:class:`ServiceChannel`): Service channel
        """
        msg_type = msg.header.flags.msg_type
        payload = msg.protected_payload
        if msg_type == MessageType.AuxilaryStream:
            if payload.connection_info_flag == 0:
                log.debug('Received AuxiliaryStream HELLO')
                self._request_connection_info()
            else:
                log.debug('Received AuxiliaryStream CONNECTION INFO')
                self.connection_info = payload.connection_info

        elif msg_type == MessageType.ActiveSurfaceChange:
            self.active_surface = payload

        else:
            raise TitleManagerError('Unhandled Msg: {}, Payload: {}'.format(msg_type, payload))

    def _request_connection_info(self):
        msg = factory.title_auxiliary_stream()
        return self._send_message(msg)

    def start_title_channel(self, title_id):
        return self.console.protocol.start_channel(ServiceChannel.Title,
                                                   MessageTarget.TitleUUID,
                                                   title_id=title_id)

    @property
    def active_surface(self):
        """
        Get `Active Surface`.

        Returns:
             :class:`XStructObj`: Active Surface
        """
        return self._active_surface

    @active_surface.setter
    def active_surface(self, value):
        """
        Set `Active Surface`.

        Args:
            value (:class:`XStructObj`): Active Surface payload

        Returns: None
        """
        self._active_surface = value
        self.on_surface_change(value)

    @property
    def connection_info(self):
        """
        Get current `Connection info`

        Returns:
            :class:`XStructObj`: Connection info
        """
        return self._connection_info

    @connection_info.setter
    def connection_info(self, value):
        """
        Set `Connection info` and setup `Crypto`-context

        Args:
            value (:class:`XStructObj`): Connection info

        Returns: None
        """
        self._connection_info = value
        self.on_connection_info(value)



from flask import Flask, jsonify
from http import HTTPStatus
from xbox.webapi.authentication.manager import AuthenticationManager
from xbox.webapi.api.client import XboxLiveClient
from .routes import routes


class SmartGlassFlaskApp(Flask):
    def __init__(self, name):
        super(SmartGlassFlaskApp, self).__init__(name)

        self.console_cache = {}
        self.title_cache = {}
        self.authentication_mgr = AuthenticationManager()
        self.token_file = None
        self._xbl_client = None

    @property
    def smartglass_packetnames(self):
        return [
            'xbox-smartglass-core',
            'xbox-smartglass-nano',
            'xbox-webapi'
        ]

    @property
    def xbl_client(self):
        if self.authentication_mgr.authenticated:
            self._xbl_client = XboxLiveClient(
                userhash=self.authentication_mgr.userinfo.userhash,
                auth_token=self.authentication_mgr.xsts_token.jwt,
                xuid=self.authentication_mgr.userinfo.xuid
            )
        return self._xbl_client

    @property
    def logged_in_gamertag(self):
        return self.authentication_mgr.userinfo.gamertag if self.authentication_mgr.userinfo else '<UNKNOWN>'

    def reset_authentication(self):
        self.authentication_mgr = AuthenticationManager()

    def error(self, message, code=HTTPStatus.INTERNAL_SERVER_ERROR, **kwargs):
        ret = {
            'success': False,
            'message': message
        }
        if kwargs:
            ret.update(kwargs)
        self.logger.error(str(ret))
        return jsonify(ret), code

    @staticmethod
    def success(**kwargs):
        ret = {'success': True}
        if kwargs:
            ret.update(kwargs)
        return jsonify(ret)


app = SmartGlassFlaskApp(__name__)
app.register_blueprint(routes)



import os
import pytest
import uuid
import json

from binascii import unhexlify
from construct import Container

from xbox.sg import enum, packer, packet

from xbox.sg.console import Console
from xbox.sg.crypto import Crypto
from xbox.sg.manager import MediaManager, TextManager, InputManager

from xbox.auxiliary.crypto import AuxiliaryStreamCrypto

from xbox.stump.manager import StumpManager

from xbox.rest.app import app as rest_app
from xbox.rest.consolewrap import ConsoleWrap


@pytest.fixture(scope='session')
def uuid_dummy():
    return uuid.UUID('de305d54-75b4-431b-adb2-eb6b9e546014')


@pytest.fixture(scope='session')
def console_address():
    return '10.11.12.12'


@pytest.fixture(scope='session')
def console_name():
    return 'TestConsole'


@pytest.fixture(scope='session')
def console_liveid():
    return 'FD0000123456789'


@pytest.fixture(scope='session')
def console_flags():
    return enum.PrimaryDeviceFlag.AllowAnonymousUsers | enum.PrimaryDeviceFlag.AllowAuthenticatedUsers


@pytest.fixture(scope='session')
def public_key_bytes():
    return unhexlify(
        b'041815d5382df79bd792a8d8342fbc717eacef6a258f779279e5463573e06b'
        b'f84c6a88fac904870bf3a26f856e65f483195c4323eef47a048f23a031da6bd0929d'
    )


@pytest.fixture(scope='session')
def shared_secret_bytes():
    return unhexlify(
        '82bba514e6d19521114940bd65121af234c53654a8e67add7710b3725db44f77'
        '30ed8e3da7015a09fe0f08e9bef3853c0506327eb77c9951769d923d863a2f5e'
    )


@pytest.fixture(scope='session')
def crypto(shared_secret_bytes):
    return Crypto.from_shared_secret(shared_secret_bytes)


@pytest.fixture(scope='session')
def console(console_address, console_name, uuid_dummy, console_liveid, console_flags, public_key_bytes):
    c = Crypto.from_bytes(public_key_bytes)
    console = Console(
        console_address, console_name, uuid_dummy,
        console_liveid, console_flags, c.foreign_pubkey
    )
    console.add_manager(StumpManager)
    console.add_manager(MediaManager)
    console.add_manager(TextManager)
    console.add_manager(InputManager)
    return console


@pytest.fixture(scope='session')
def public_key(public_key_bytes):
    c = Crypto.from_bytes(public_key_bytes)
    return c.foreign_pubkey


@pytest.fixture(scope='session')
def packets():
    # Who cares about RAM anyway?
    data = {}
    data_path = os.path.join(os.path.dirname(__file__), 'data', 'packets')
    for f in os.listdir(data_path):
        with open(os.path.join(data_path, f), 'rb') as fh:
            data[f] = fh.read()

    return data


@pytest.fixture(scope='session')
def stump_json():
    # Who cares about RAM anyway?
    data = {}
    data_path = os.path.join(os.path.dirname(__file__), 'data', 'stump_json')
    for f in os.listdir(data_path):
        with open(os.path.join(data_path, f), 'rt') as fh:
            data[f] = json.load(fh)

    return data


@pytest.fixture(scope='session')
def decrypted_packets(packets, crypto):
    return {k: packer.unpack(v, crypto) for k, v in packets.items()}


@pytest.fixture(scope='session')
def pcap_filepath():
    return os.path.join(os.path.dirname(__file__), 'data', 'sg_capture.pcap')


@pytest.fixture(scope='session')
def certificate_data():
    filepath = os.path.join(os.path.dirname(__file__), 'data', 'selfsigned_cert.bin')
    with open(filepath, 'rb') as f:
        data = f.read()
    return data


@pytest.fixture(scope='session')
def json_fragments():
    filepath = os.path.join(os.path.dirname(__file__), 'data', 'json_fragments.json')
    with open(filepath, 'rt') as f:
        data = json.load(f)
    return data['fragments']


@pytest.fixture(scope='session')
def aux_streams():
    data = {}
    data_path = os.path.join(os.path.dirname(__file__), 'data', 'aux_streams')
    for f in os.listdir(data_path):
        with open(os.path.join(data_path, f), 'rb') as fh:
            data[f] = fh.read()

    return data


@pytest.fixture(scope='session')
def aux_crypto(decrypted_packets):
    connection_info = decrypted_packets['auxiliary_stream_connection_info'].protected_payload.connection_info
    return AuxiliaryStreamCrypto.from_connection_info(connection_info)


@pytest.fixture
def rest_client():
    rest_app.config['TESTING'] = True
    yield rest_app


@pytest.fixture(scope='session')
def media_state():
    return packet.message.media_state(
        title_id=274278798,
        aum_id='AIVDE_s9eep9cpjhg6g!App',
        asset_id='',
        media_type=enum.MediaType.Video,
        sound_level=enum.SoundLevel.Full,
        enabled_commands=enum.MediaControlCommand.Play | enum.MediaControlCommand.Pause,
        playback_status=enum.MediaPlaybackStatus.Playing,
        rate=1.00,
        position=0,
        media_start=0,
        media_end=0,
        min_seek=0,
        max_seek=0,
        metadata=[
            Container(name='title', value='Some Movietitle'),
            Container(name='subtitle', value='')
        ]
    )


@pytest.fixture(scope='session')
def active_title():
    struct = packet.message._active_title(
        title_id=714681658,
        product_id=uuid.UUID('00000000-0000-0000-0000-000000000000'),
        sandbox_id=uuid.UUID('00000000-0000-0000-0000-000000000000'),
        aum='Xbox.Home_8wekyb3d8bbwe!Xbox.Home.Application',
        disposition=Container(
            has_focus=True,
            title_location=enum.ActiveTitleLocation.StartView
        )
    )
    return struct


@pytest.fixture(scope='session')
def active_media_title():
    struct = packet.message._active_title(
        title_id=714681658,
        product_id=uuid.UUID('00000000-0000-0000-0000-000000000000'),
        sandbox_id=uuid.UUID('00000000-0000-0000-0000-000000000000'),
        aum='AIVDE_s9eep9cpjhg6g!App',
        disposition=Container(
            has_focus=True,
            title_location=enum.ActiveTitleLocation.StartView
        )
    )
    return struct


@pytest.fixture(scope='session')
def console_status(active_title):
    return packet.message.console_status(
        live_tv_provider=0,
        major_version=10,
        minor_version=0,
        build_number=14393,
        locale='en-US',
        active_titles=[
            active_title
        ]
    )


@pytest.fixture(scope='session')
def console_status_with_media(active_media_title):
    return packet.message.console_status(
        live_tv_provider=0,
        major_version=10,
        minor_version=0,
        build_number=14393,
        locale='en-US',
        active_titles=[
            active_media_title
        ]
    )


@pytest.fixture
def rest_client_connected_media_console_status(rest_client, console, media_state, console_status_with_media):
    console._device_status = enum.DeviceStatus.Available
    console._connection_state = enum.ConnectionState.Connected
    console._pairing_state = enum.PairedIdentityState.Paired

    console.media._media_state = media_state
    console._console_status = console_status_with_media

    console_wrap = ConsoleWrap(console)
    rest_client.console_cache[console.liveid] = console_wrap
    return rest_client



"""
Encrypt smartglass messages (type 0xD00D) with a new key

Example:
    usage: xbox_recrypt [-h] src_path src_secret dst_path dst_secret

    Re-Encrypt raw smartglass packets from a given filepath

    positional arguments:
      src_path    Path to sourcefiles
      src_secret  Source shared secret in hex-format
      dst_path    Path to destination
      dst_secret  Target shared secret in hex-format

    optional arguments:
      -h, --help  show this help message and exit
"""
import os
import sys
import argparse
from binascii import unhexlify

from construct import Int16ub

from xbox.sg.crypto import Crypto
from xbox.sg.enum import PacketType


def main():
    parser = argparse.ArgumentParser(
        description='Re-Encrypt raw smartglass packets from a given filepath'
    )
    parser.add_argument('src_path', type=str, help='Path to sourcefiles')
    parser.add_argument('src_secret', type=str, help='Source shared secret in hex-format')
    parser.add_argument('dst_path', type=str, help='Path to destination')
    parser.add_argument('dst_secret', type=str, help='Target shared secret in hex-format')
    args = parser.parse_args()

    src_secret = unhexlify(args.src_secret)
    dst_secret = unhexlify(args.dst_secret)

    src_path = args.src_path
    dst_path = args.dst_path

    if len(src_secret) != 64:
        print('Source key of invalid length supplied!')
        sys.exit(1)
    elif len(dst_secret) != 64:
        print('Destination key of invalid length supplied!')
        sys.exit(1)

    source_crypto = Crypto.from_shared_secret(src_secret)
    dest_crypto = Crypto.from_shared_secret(dst_secret)

    source_path = os.path.dirname(src_path)
    dest_path = os.path.dirname(dst_path)

    for f in os.listdir(source_path):
        src_filepath = os.path.join(source_path, f)
        with open(src_filepath, 'rb') as sfh:
            encrypted = sfh.read()
            if Int16ub.parse(encrypted[:2]) != PacketType.Message.value:
                print('Invalid magic, %s not a smartglass message, ignoring'
                      % src_filepath)
                continue

            # Slice the encrypted data manually
            header = encrypted[:26]
            payload = encrypted[26:-32]
            hash = encrypted[-32:]

            if not source_crypto.verify(encrypted[:-32], hash):
                print('Hash mismatch, ignoring')
                continue

            # Decrypt with source shared secret
            iv = source_crypto.generate_iv(header[:16])
            decrypted_payload = source_crypto.decrypt(iv, payload)

            # Encrypt with destination parameters
            new_iv = dest_crypto.generate_iv(header[:16])
            recrypted = dest_crypto.encrypt(new_iv, decrypted_payload)
            new_hash = dest_crypto.hash(header + recrypted)
            new_packet = header + recrypted + new_hash

            with open(os.path.join(dest_path, f), 'wb') as dfh:
                dfh.write(new_packet)


if __name__ == '__main__':
    main()



from xbox.sg import console
from xbox.sg import enum
from xbox.sg import packer


def test_init(public_key, uuid_dummy):
    c = console.Console(
        '10.0.0.23', 'XboxOne', uuid_dummy, 'FFFFFFFFFFF',
        enum.PrimaryDeviceFlag.AllowConsoleUsers, 0, public_key
    )

    assert c.address == '10.0.0.23'
    assert c.flags == enum.PrimaryDeviceFlag.AllowConsoleUsers
    assert c.name == 'XboxOne'
    assert c.uuid == uuid_dummy
    assert c.liveid == 'FFFFFFFFFFF'
    assert c._public_key is not None
    assert c._crypto is not None
    assert c.device_status == enum.DeviceStatus.Unavailable
    assert c.connection_state == enum.ConnectionState.Disconnected
    assert c.pairing_state == enum.PairedIdentityState.NotPaired
    assert c.paired is False
    assert c.available is False
    assert c.connected is False

    assert c.authenticated_users_allowed is False
    assert c.anonymous_connection_allowed is False
    assert c.console_users_allowed is True
    assert c.is_certificate_pending is False


def test_init_from_message(packets, crypto, uuid_dummy):
    msg = packer.unpack(packets['discovery_response'], crypto)

    c = console.Console.from_message('10.0.0.23', msg)

    assert c.address == '10.0.0.23'
    assert c.flags == enum.PrimaryDeviceFlag.AllowAuthenticatedUsers
    assert c.name == 'XboxOne'
    assert c.uuid == uuid_dummy
    assert c.liveid == 'FFFFFFFFFFF'
    assert c._public_key is not None
    assert c._crypto is not None
    assert c.device_status == enum.DeviceStatus.Available
    assert c.connection_state == enum.ConnectionState.Disconnected
    assert c.pairing_state == enum.PairedIdentityState.NotPaired
    assert c.paired is False
    assert c.available is True
    assert c.connected is False

    assert c.authenticated_users_allowed is True
    assert c.anonymous_connection_allowed is False
    assert c.console_users_allowed is False
    assert c.is_certificate_pending is False



from flask import current_app as app
from . import routes


@routes.route('/versions')
def library_versions():
    import pkg_resources

    versions = {}
    for name in app.smartglass_packetnames:
        try:
            versions[name] = pkg_resources.get_distribution(name).version
        except Exception:
            versions[name] = None

    return app.success(versions=versions)



from flask import current_app as app
from http import HTTPStatus
from functools import wraps

"""
Decorators
"""


def console_connected(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        liveid = kwargs.get('liveid')
        console = app.console_cache.get(liveid)
        if not console:
            return app.error('Console {0} is not alive'.format(liveid), HTTPStatus.FORBIDDEN)
        elif not console.connected:
            return app.error('Console {0} is not connected'.format(liveid), HTTPStatus.FORBIDDEN)

        del kwargs['liveid']
        kwargs['console'] = console
        return f(*args, **kwargs)
    return decorated_function


def console_exists(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        liveid = kwargs.get('liveid')
        console = app.console_cache.get(liveid)
        if not console:
            return app.error('Console info for {0} is not available'.format(liveid), HTTPStatus.FORBIDDEN)

        del kwargs['liveid']
        kwargs['console'] = console
        return f(*args, **kwargs)
    return decorated_function


def require_authentication(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        if not app.authentication_mgr.authenticated:
            return app.error('Not authenticated for Xbox Live', HTTPStatus.UNAUTHORIZED)

        kwargs['client'] = app.xbl_client
        return f(*args, **kwargs)
    return decorated_function



from flask import Blueprint
routes = Blueprint('routes', __name__)

from .index import *  # noqa: F401,F403
from .auth import *  # noqa: F401,F403
from .device import *  # noqa: F401,F403
from .web import *  # noqa: F401,F403
from .versions import *  # noqa: F401,F403



from http import HTTPStatus


def test_api_overview(rest_client):
    resp = rest_client.test_client().get('/')

    assert resp.status_code == HTTPStatus.OK
    assert resp.json['success'] is True

    assert 'endpoints' in resp.json
    assert len(resp.json['endpoints']) > 1



"""
Fallout 4 AuxiliaryStream Relay client

Tunnels packets via TCP/27000 - compatible with regular PipBoy-clients
"""
import logging
from xbox.auxiliary.relay import AuxiliaryRelayService

LOGGER = logging.getLogger(__name__)

FALLOUT_TITLE_ID = 0x4ae8f9b2
LOCAL_PIPBOY_PORT = 27000


def on_connection_info(info):
    print('Setting up relay on TCP/{0}...\n'.format(LOCAL_PIPBOY_PORT))
    service = AuxiliaryRelayService(info, listen_port=LOCAL_PIPBOY_PORT)
    service.run()



"""
Smartglass protocol core

**NOTE**: Should not be used directly, use :class:`.Console` !
"""

import uuid
import json
import base64
import logging
import gevent
import gevent.event
from gevent import socket
from gevent.server import DatagramServer

from xbox.sg import factory, packer
from xbox.sg.packet.message import message_structs
from xbox.sg.enum import PacketType, ConnectionResult, DisconnectReason,\
    ServiceChannel, MessageType, AckStatus, SGResultCode, ActiveTitleLocation
from xbox.sg.constants import WindowsClientInfo, MessageTarget
from xbox.sg.utils.events import Event

log = logging.getLogger(__name__)

PORT = 5050
BROADCAST = '<broadcast>'
MULTICAST = '239.255.255.250'

CHANNEL_MAP = {
    ServiceChannel.SystemInput: MessageTarget.SystemInputUUID,
    ServiceChannel.SystemInputTVRemote: MessageTarget.SystemInputTVRemoteUUID,
    ServiceChannel.SystemMedia: MessageTarget.SystemMediaUUID,
    ServiceChannel.SystemText: MessageTarget.SystemTextUUID,
    ServiceChannel.SystemBroadcast: MessageTarget.SystemBroadcastUUID
}


class ProtocolError(Exception):
    """
    Exception thrown by CoreProtocol
    """
    pass


class CoreProtocol(DatagramServer):
    HEARTBEAT_INTERVAL = 3.0

    def __init__(self, address=None, crypto=None):
        """
        Instantiate Smartglass Protocol handler.

        Args:
            address (str): Optional IP address to send all packets to
            crypto (:class:`.Crypto`): Instance of crypto context
        """
        self.addr = address
        self.crypto = crypto

        self._discovered = {}

        self.target_participant_id = None
        self.source_participant_id = None

        self._pending = {}
        self._chl_mgr = ChannelManager()
        self._seq_mgr = SequenceManager()
        self._frg_mgr = FragmentManager()

        self.on_timeout = Event()
        self.on_discover = Event()
        self.on_message = Event()
        self.on_json = Event()

        super(DatagramServer, self).__init__(('*:0'))

    @classmethod
    def get_listener(cls, *args, **kwargs):
        sock = DatagramServer.get_listener(*args, **kwargs)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        return sock

    def stop(self, *args, **kwargs):
        """
        Stop protocol.

        Disconnect from console, then stop the protocol handler.

        Args:
            *args: Args
            **kwargs: KwArgs

        Returns: None
        """
        if self.started:
            try:
                self.disconnect()  # Best effort disconnect
            except Exception:
                pass
            super(DatagramServer, self).stop(*args, **kwargs)

    def send_message(self, msg, channel=ServiceChannel.Core, addr=None,
                     blocking=True, timeout=5, retries=3):
        """
        Send message to console.

        Packing and encryption happens here.

        Args:
            msg (:obj:`XStruct`): Unassembled message to send
            channel (:class:`ServiceChannel`): Channel to send the message on,
                           Enum member of `ServiceChannel`
            addr (str): IP address of target console
            blocking (bool): If set and `msg` is `Message`-packet, wait for ack
            timeout (int): Seconds to wait for ack, only useful if `blocking`
                           is `True`
            retries (int): Max retry count.

        Returns: None

        Raises:
            ProtocolError: On failure
        """
        if msg.header.pkt_type == PacketType.Message:
            msg.header(
                sequence_number=self._seq_mgr.next_sequence_num(),
                target_participant_id=self.target_participant_id,
                source_participant_id=self.source_participant_id,
                channel_id=self._chl_mgr.get_channel_id(channel)
            )

        if self.addr:
            addr = self.addr

        if self.crypto:
            data = packer.pack(msg, self.crypto)
        else:
            data = packer.pack(msg)

        if not addr:
            raise ProtocolError("No address specified in send_message")

        if not data:
            raise ProtocolError("No data")

        if msg.header.pkt_type == PacketType.Message \
                and msg.header.flags.need_ack and blocking:
            log.debug(
                "Sending %s message on ServiceChannel %s to %s",
                msg.header.flags.msg_type.name, channel.name, addr,
                extra={'_msg': msg}
            )
            seqn = msg.header.sequence_number
            tries = 0
            result = None

            while tries < retries and not result:
                if tries > 0:
                    log.warning(
                        "Message %s on ServiceChannel %s to %s not ack'd in time, attempt #%d",
                        msg.header.flags.msg_type.name, channel.name, addr, tries + 1,
                        extra={'_msg': msg}
                    )

                self.socket.sendto(data, (addr, PORT))
                result = self._await_ack('ack_%i' % seqn, timeout)
                tries += 1

            if result:
                return result

            raise ProtocolError("Exceeded retries")
        elif msg.header.pkt_type == PacketType.ConnectRequest:
            log.debug("Sending ConnectRequest to %s", addr, extra={'_msg': msg})

        self.socket.sendto(data, (addr, PORT))

    def handle(self, data, addr):
        """
        Handle incoming smartglass packets

        Args:
            data (bytes): Raw packet
            addr (str): IP address of sender

        Returns: None
        """
        try:
            host, _ = addr

            if self.crypto:
                msg = packer.unpack(data, self.crypto)
            else:
                msg = packer.unpack(data)

            if msg.header.pkt_type == PacketType.DiscoveryResponse:
                log.debug("Received DiscoverResponse from %s", host, extra={'_msg': msg})
                self._discovered[host] = msg
                self.on_discover(host, msg)

            elif msg.header.pkt_type == PacketType.ConnectResponse:
                log.debug("Received ConnectResponse from %s", host, extra={'_msg': msg})
                if 'connect' in self._pending:
                    self._set_result('connect', msg)

            elif msg.header.pkt_type == PacketType.Message:
                channel = self._chl_mgr.get_channel(msg.header.channel_id)
                message_info = msg.header.flags.msg_type.name

                if msg.header.flags.is_fragment:
                    message_info = 'MessageFragment ({0})'.format(message_info)

                log.debug(
                    "Received %s message on ServiceChannel %s from %s",
                    message_info, channel.name, host, extra={'_msg': msg}
                )
                seq_num = msg.header.sequence_number
                self._seq_mgr.add_received(seq_num)

                if msg.header.flags.need_ack:
                    self.ack(
                        [msg.header.sequence_number], [], ServiceChannel.Core
                    )

                self._seq_mgr.low_watermark = seq_num

                if msg.header.flags.is_fragment:
                    sequence_begin = msg.protected_payload.sequence_begin
                    sequence_end = msg.protected_payload.sequence_end
                    fragment_payload = self._frg_mgr.reassemble_message(msg)
                    if not fragment_payload:
                        return

                    msg(protected_payload=fragment_payload)
                    log.debug("Assembled {0} (Seq {1}:{2})".format(
                        message_info, sequence_begin, sequence_end
                    ), extra={'_msg': msg})

                self._on_message(msg, channel)
            else:
                self._on_unk(msg)
        except Exception:
            log.exception("Exception in CoreProtocol datagram handler")

    def _await_ack(self, identifier, timeout=5):
        """
        Wait for acknowledgement of message

        Args:
            identifier (str): Identifier of ack
            timeout (int): Timeout in seconds

        Returns:
            :obj:`.Event`: Event
        """
        evnt = gevent.event.AsyncResult()
        self._pending[identifier] = evnt

        return evnt.wait(timeout)

    def _set_result(self, identifier, result):
        """
        Called when an acknowledgement comes in, unblocks `_await_ack`

        Args:
            identifier (str): Identifier of ack
            result (int): Ack status

        Returns: None
        """
        self._pending[identifier].set(result)
        del self._pending[identifier]

    def _heartbeat(self):
        """
        Greenlet checking for console activity, firing `on_timeout`-event on timeout.

        Heartbeats are empty "ack" messages that are to be ack'd by the console.

        Returns:
            None
        """
        while self.started:
            try:
                self.ack([], [], ServiceChannel.Core, need_ack=True)
            except ProtocolError:
                self.on_timeout()
                self.stop()
                break
            gevent.sleep(self.HEARTBEAT_INTERVAL)

    def _on_message(self, msg, channel):
        """
        Handle msg of type `Message`.

        Args:
            msg (:class:`.XStruct`): Message
            channel (:class:`ServiceChannel`): Channel the message was received on

        Returns: None
        """
        msg_type = msg.header.flags.msg_type

        # First run our internal handlers
        if msg_type == MessageType.Ack:
            self._on_ack(msg)

        elif msg_type == MessageType.StartChannelResponse:
            self._chl_mgr.handle_channel_start_response(msg)

        elif msg_type == MessageType.Json:
            self._on_json(msg, channel)

        # Then our hooked handlers
        self.on_message(msg, channel)

    def _on_ack(self, msg):
        """
        Process acknowledgement message.

        Args:
            msg (:class:`.XStruct`): Message

        Returns: None
        """

        for num in msg.protected_payload.processed_list:
            identifier = 'ack_%i' % num
            self._seq_mgr.add_processed(num)
            if identifier in self._pending:
                self._set_result(identifier, AckStatus.Processed)
        for num in msg.protected_payload.rejected_list:
            identifier = 'ack_%i' % num
            self._seq_mgr.add_rejected(num)
            if identifier in self._pending:
                self._set_result(identifier, AckStatus.Rejected)

    def _on_json(self, msg, channel):
        """
        Process json message.

        Args:
            msg (:class:`.XStruct`): Message
            channel (:class:`ServiceChannel`): Channel the message was received on

        Returns: None
        """
        text = msg.protected_payload.text

        if 'fragment_data' in text:
            text = self._frg_mgr.reassemble_json(text)
            if not text:
                # Input message is a fragment, but cannot assemble full msg yet
                return

        self.on_json(text, channel)

    def discover(self, addr=None, tries=5, blocking=True, timeout=5):
        """
        Discover consoles on the network

        Args:
            addr (str): IP address
            tries (int): Discover attempts
            blocking (bool): Wait a given time for responses, otherwise
                             return immediately
            timeout (int): Timeout in seconds (only if `blocking` is `True`)

        Returns:
            list: List of discovered consoles
        """
        self._discovered = {}
        msg = factory.discovery()

        greenlet = gevent.spawn(self._discover, msg, addr, tries)

        # Blocking for a discovery is different than connect or regular message
        if blocking:
            with gevent.Timeout(timeout, False):
                greenlet.join()

        return self.discovered

    def _discover(self, msg, addr, tries):
        for _ in range(tries):
            self.send_message(msg, addr=BROADCAST)
            self.send_message(msg, addr=MULTICAST)

            if addr:
                self.send_message(msg, addr=addr)

            gevent.sleep(0.5)

    @property
    def discovered(self):
        """
        Return discovered consoles

        Returns:
            dict: Discovered consoles
        """
        return self._discovered

    def connect(self, userhash, xsts_token, client_uuid=uuid.uuid4(),
                request_num=0, retries=3):
        """
        Connect to console

        Args:
            userhash (str): Userhash from Xbox Live Authentication
            xsts_token (str): XSTS Token from Xbox Live Authentication
            client_uuid (UUID): Client UUID (default: Generate random uuid)
            request_num (int): Request number
            retries (int): Max. connect attempts

        Returns:
            int: Pairing State

        Raises:
            ProtocolError: If connection fails
        """
        if not self.crypto:
            raise ProtocolError("No crypto")

        if isinstance(userhash, type(None)):
            userhash = ''
        if isinstance(xsts_token, type(None)):
            xsts_token = ''

        iv = self.crypto.generate_iv()
        pubkey_type = self.crypto.pubkey_type
        pubkey = self.crypto.pubkey_bytes

        msg = factory.connect(
            client_uuid, pubkey_type, pubkey, iv, userhash, xsts_token,
            request_num, request_num, request_num + 1
        )

        payload_len = packer.payload_length(msg)
        if payload_len < 1024:
            messages = [msg]
        else:
            messages = _fragment_connect_request(
                self.crypto, client_uuid, pubkey_type, pubkey,
                userhash, xsts_token, request_num
            )

        tries = 0
        result = None
        while tries < retries and not result:
            for m in messages:
                self.send_message(m)

            result = self._await_ack('connect')

        if not result:
            raise ProtocolError("Exceeded connect retries")

        connect_result = result.protected_payload.connect_result
        if connect_result != ConnectionResult.Success:
            raise ProtocolError(
                "Connecting failed! Result: %s" % connect_result
            )

        self.target_participant_id = 0
        self.source_participant_id = result.protected_payload.participant_id

        self.local_join()

        for channel, target_uuid in CHANNEL_MAP.items():
            self.start_channel(channel, target_uuid)

        gevent.spawn_later(self.HEARTBEAT_INTERVAL, self._heartbeat)
        return result.protected_payload.pairing_state

    def local_join(self, client_info=WindowsClientInfo, **kwargs):
        """
        Pair client with console.

        Args:
            client_info (`ClientInfo`): Either `WindowsClientInfo` or
                                        `AndroidClientInfo`
            **kwargs:

        Returns: None
        """
        msg = factory.local_join(client_info)
        return self.send_message(msg, **kwargs)

    def start_channel(self, channel, messagetarget_uuid, title_id=0, activity_id=0, **kwargs):
        """
        Request opening of specific ServiceChannel

        Args:
            channel (:class:`ServiceChannel`): Channel to start
            messagetarget_uuid (UUID): Message Target UUID
            title_id (int): Title ID, Only used for ServiceChannel.Title
            activity_id (int): Activity ID, unknown use-case
            **kwargs: KwArgs

        Returns: None
        """
        request_id = self._chl_mgr.get_next_request_id(channel)
        msg = factory.start_channel(request_id, title_id, messagetarget_uuid, activity_id)
        return self.send_message(msg, **kwargs)

    def ack(self, processed, rejected, channel, need_ack=False):
        """
        Acknowledge received messages that have `need_ack` flag set.

        Args:
            processed (list): Processed sequence numbers
            rejected (list): Rejected sequence numbers
            channel (:class:`ServiceChannel`): Channel to send the ack on
            need_ack (bool): Whether we want this ack to be ack'd. Will be blocking if set.
                             Required for heartbeat messages.

        Returns: None
        """
        low_watermark = self._seq_mgr.low_watermark
        msg = factory.acknowledge(
            low_watermark, processed, rejected, need_ack=need_ack
        )
        self.send_message(msg, channel=channel, blocking=need_ack)

    def json(self, data, channel):
        """
        Send json message

        Args:
            data (dict): JSON dict
            channel (:class:`ServiceChannel`): Channel to send the message to

        Returns: None
        """
        msg = factory.json(data)
        self.send_message(msg, channel=channel)

    def power_on(self, liveid, addr=None, tries=2):
        """
        Power on console.

        Args:
            liveid (str): Live ID of console
            addr (str): (optional) IP address of console
            tries (int): PowerOn attempts

        Returns: None
        """
        msg = factory.power_on(liveid)

        for i in range(tries):
            self.send_message(msg, addr=BROADCAST)
            self.send_message(msg, addr=MULTICAST)
            if addr:
                self.send_message(msg, addr=addr)

            gevent.sleep(0.1)

    def power_off(self, liveid):
        """
        Power off console

        Args:
            liveid (str): Live ID of console

        Returns: None
        """
        msg = factory.power_off(liveid)
        self.send_message(msg)

    def disconnect(self, reason=DisconnectReason.Unspecified, error=0):
        """
        Disconnect console session

        Args:
            reason (:class:`DisconnectReason`): Disconnect reason
            error (int): Error Code

        Returns: None
        """
        msg = factory.disconnect(reason, error)
        self.send_message(msg)

    def game_dvr_record(self, start_delta, end_delta):
        """
        Start Game DVR recording

        Args:
            start_delta (int): Start time
            end_delta (int): End time

        Returns:
            int: Member of :class:`AckStatus`
        """
        msg = factory.game_dvr_record(start_delta, end_delta)
        return self.send_message(msg)

    def launch_title(self, uri, location=ActiveTitleLocation.Full):
        """
        Launch title via URI

        Args:
            uri (str): Uri string
            location (:class:`ActiveTitleLocation`): Location

        Returns:
            int: Member of :class:`AckStatus`
        """
        msg = factory.title_launch(location, uri)
        return self.send_message(msg)


class SequenceManager(object):
    def __init__(self):
        """
        Process received messages by sequence numbers.
        Also add processed / rejected messages to a list.
        Tracks the `Low Watermark` that's sent with
        `Acknowledgement`-Messages too.
        """
        self.processed = []
        self.rejected = []
        self.received = []

        self._low_watermark = 0
        self._sequence_num = 0

    def add_received(self, num):
        """
        Add received sequence number

        Args:
            num (int): Sequence number

        Returns: None
        """
        if num not in self.received:
            self.received.append(num)

    def add_processed(self, num):
        """
        Add sequence number of message that was sent to console
        and succeeded in processing.

        Args:
            num (int): Sequence number

        Returns: None
        """
        if num not in self.processed:
            self.processed.append(num)

    def add_rejected(self, num):
        """
        Add sequence number of message that was sent to console
        and was rejected by it.

        Args:
            num (int): Sequence number

        Returns: None
        """
        if num not in self.rejected:
            self.rejected.append(num)

    def next_sequence_num(self):
        """
        Get next sequence number to use for outbound `Message`.

        Returns: None
        """
        self._sequence_num += 1
        return self._sequence_num

    @property
    def low_watermark(self):
        """
        Get current `Low Watermark`

        Returns:
            int: Low Watermark
        """
        return self._low_watermark

    @low_watermark.setter
    def low_watermark(self, value):
        """
        Set `Low Watermark`

        Args:
            value (int): Last received sequence number from console

        Returns: None
        """
        if value > self._low_watermark:
            self._low_watermark = value


class ChannelError(Exception):
    """
    Exception thrown by :class:`ChannelManager`.
    """
    pass


class ChannelManager(object):
    CHANNEL_CORE = 0
    CHANNEL_ACK = 0x1000000000000000

    def __init__(self):
        """
        Keep track of established ServiceChannels
        """
        self._channel_mapping = {}
        self._requests = {}
        self._request_id = 0

    def handle_channel_start_response(self, msg):
        """
        Handle message of type `StartChannelResponse`

        Args:
            msg (:class:`XStructObj`): Start Channel Response message

        Raises:
            :class:`ChannelError`: If channel acquire failed

        Returns:
            :class:`ServiceChannel`: Acquired ServiceChannel
        """
        # Find ServiceChannel by RequestId
        request_id = msg.protected_payload.channel_request_id
        channel = self._requests.get(request_id)

        if not channel:
            raise ChannelError("Request Id %d not found. Was the channel request saved?" % request_id)

        if msg.protected_payload.result != SGResultCode.SG_E_SUCCESS:
            raise ChannelError("Acquiring ServiceChannel %s failed" % channel.name)

        # Save Channel Id for appropriate ServiceChannel
        channel_id = msg.protected_payload.target_channel_id
        self._channel_mapping[channel] = channel_id

        self._requests.pop(request_id)
        log.debug("Acquired ServiceChannel %s -> Channel: 0x%x", channel.name, channel_id)
        return channel

    def get_next_request_id(self, channel):
        """
        Get next Channel request id for ServiceChannel

        Incremented on each call.

        Args:
            channel (:class:`ServiceChannel`): Service channel

        Returns:
            int: Channel request id
        """
        # Clear old request for same ServiceChannel
        self._requests = {key: val for key, val in self._requests.items()
                          if val != channel}

        self._request_id += 1
        self._requests[self._request_id] = channel
        return self._request_id

    def get_channel(self, channel_id):
        """
        Get matching :class:`ServiceChannel` for provided Channel ID of `Message`

        Args:
            channel_id (int): Channel of Message

        Returns:
            :class:`ServiceChannel`: Service channel
        """
        # Core and Ack are fixed, don't need mapping
        if channel_id == self.CHANNEL_CORE:
            return ServiceChannel.Core
        elif channel_id == self.CHANNEL_ACK:
            return ServiceChannel.Ack

        for key, value in self._channel_mapping.items():
            if value == channel_id:
                return key
        raise ChannelError("ServiceChannel not found for channel_id: 0x%x"
                           % channel_id)

    def get_channel_id(self, channel):
        """
        Get Channel ID for use in `Message` for provided :class:`ServiceChannel`

        Args:
            channel (:class:`ServiceChannel`): Service channel

        Returns:
            int: Channel ID for use in `Message`
        """

        # Core and Ack are fixed, don't need mapping
        if channel == ServiceChannel.Core:
            return self.CHANNEL_CORE
        elif channel == ServiceChannel.Ack:
            return self.CHANNEL_ACK

        if channel not in self._channel_mapping:
            raise ChannelError("Channel ID not found for ServiceChannel: %s" % channel)
        return self._channel_mapping[channel]

    def reset(self):
        """
        Erase the channels table

        Returns:
            None
        """
        self._requests = {}
        self._channel_mapping = {}
        self._request_id = 0


class FragmentError(Exception):
    """
    Exception thrown by :class:`FragmentManager`.
    """
    pass


class FragmentManager(object):
    """
    Assembles fragmented messages
    """
    def __init__(self):
        self.msg_queue = {}
        self.json_queue = {}

    def reassemble_message(self, msg):
        """
        Reassemble message fragment

        Args:
            msg (:class:`.XStruct`): Message fragment

        Returns:
            :class:`.XStruct`: Reassembled / decoded payload on success,
                `None` if payload is not ready or assembly failed
        """
        msg_type = msg.header.flags.msg_type
        payload = msg.protected_payload
        current_sequence = msg.header.sequence_number
        sequence_begin = payload.sequence_begin
        sequence_end = payload.sequence_end

        self.msg_queue[current_sequence] = payload.data

        wanted_sequences = list(range(sequence_begin, sequence_end))
        assembled = b''
        for s in wanted_sequences:
            data = self.msg_queue.get(s)
            if not data:
                return

            assembled += data

        [self.msg_queue.pop(s) for s in wanted_sequences]

        # Parse raw data with original message struct
        struct = message_structs.get(msg_type)
        if not struct:
            raise FragmentError(
                'Failed to find message struct for fragmented {0}'.format(msg_type)
            )

        return struct.parse(assembled)

    def reassemble_json(self, json_msg):
        """
        Reassemble fragmented json message

        Args:
            json_msg (dict): Fragmented json message

        Returns:
            str: Reassembled / Decoded string on success,
                 `None` if datagram is not ready or assembly failed
        """
        datagram_id, datagram_size = int(json_msg['datagram_id']), int(json_msg['datagram_size'])
        fragment_offset = int(json_msg['fragment_offset'])

        fragments = self.json_queue.get(datagram_id)

        if not fragments:
            # Just add initial fragment
            self.json_queue[datagram_id] = [json_msg]
            return None

        # It's a follow-up fragment
        # Check if we already received this datagram
        for entry in fragments:
            if fragment_offset == int(entry['fragment_offset']):
                return

        # Append current fragment to datagram list
        fragments.append(json_msg)

        # Check if fragment can be assembled
        # If so, assemble and pop the fragments from queue
        if sum(int(f['fragment_length']) for f in fragments) == datagram_size:
            sorted_fragments = sorted(
                fragments, key=lambda f: int(f['fragment_offset'])
            )

            output = ''.join(f['fragment_data'] for f in sorted_fragments)

            self.json_queue.pop(datagram_id)
            return self._decode(output)

        return None

    @staticmethod
    def _encode(obj):
        """
        Dump a dict as json string, then encode with base64

        Args:
            obj (dict): Dict to encode

        Returns:
            str: base64 encoded string
        """
        bytestr = json.dumps(obj, separators=(',', ':'), sort_keys=True).encode('utf-8')
        return base64.b64encode(bytestr).decode('utf-8')

    @staticmethod
    def _decode(data):
        """
        Decode a base64 encoded json object

        Args:
            data (str): Base64 string

        Returns:
            dict: Decoded json object
        """
        return json.loads(base64.b64decode(data).decode('utf-8'))


def _fragment_connect_request(crypto, client_uuid, pubkey_type, pubkey,
                              userhash, auth_token, request_num=0):
    """
    Internal method to fragment ConnectRequest.

    Args:
        crypto (Crypto): Instance of :class:`Crypto`
        client_uuid (UUID): Client UUID
        pubkey_type (:obj:`.PublicKeyType`): Public Key Type
        pubkey (bytes): Public Key
        userhash (str): Xbox Live Account userhash
        auth_token (str): Xbox Live Account authentication token (XSTS)
        request_num (int): Request Number

    Returns:
        list: List of ConnectRequest fragments
    """
    messages = []

    # Calculate packet length (without authentication data)
    dummy_msg = factory.connect(
        client_uuid, pubkey_type, pubkey, b'\x00' * 16, u'', u'', 0, 0, 0
    )
    dummy_payload_len = packer.payload_length(dummy_msg)

    # Do fragmenting
    total_auth_len = len(userhash + auth_token)
    max_size = 1024 - dummy_payload_len

    fragments = total_auth_len // max_size
    overlap = total_auth_len % max_size

    if overlap > 0:
        fragments += 1

    group_start = request_num
    group_end = group_start + fragments

    if fragments <= 1:
        raise FragmentError('Authentication data too small to fragment')

    auth_position = 0
    for fragment_num in range(fragments):
        available = max_size
        current_hash = u''

        if fragment_num == 0:
            current_hash = userhash
            available -= len(current_hash)

        current_auth = auth_token[auth_position: auth_position + available]
        auth_position += len(current_auth)

        iv = crypto.generate_iv()
        messages.append(
            factory.connect(
                client_uuid, pubkey_type, pubkey, iv,
                current_hash, current_auth, request_num + fragment_num,
                group_start, group_end)
        )

    return messages



import construct
from xbox.sg.utils import struct, adapters


def test_xstruct():
    test_struct = struct.XStruct(
        'a' / construct.Int32ub,
        'b' / construct.Int16ub
    )

    obj = test_struct(a=1, b=2)

    assert test_struct.a.subcon == construct.Int32ub
    assert test_struct.b.subcon == construct.Int16ub
    assert 'a' in test_struct
    assert 'b' in test_struct
    assert obj.container == construct.Container(a=1, b=2)
    assert obj.a == 1
    assert obj.b == 2
    assert obj.build() == b'\x00\x00\x00\x01\x00\x02'

    obj.parse(b'\x00\x00\x00\x02\x00\x03')
    assert obj.container == construct.Container(a=2, b=3)

    obj = test_struct.parse(b'\x00\x00\x00\x03\x00\x04')
    assert obj.container == construct.Container(a=3, b=4)


def test_flatten():
    test_sub = struct.XStruct(
        'c' / construct.Int16ub
    )
    test_struct = struct.XStruct(
        'a' / construct.Int32ub,
        'b' / test_sub
    )

    obj = test_struct(a=1, b=test_sub(c=2))
    flat = struct.flatten(obj.container)

    assert flat == construct.Container(a=1, b=construct.Container(c=2))


def test_terminated_field():
    test_struct = struct.XStruct(
        'a' / adapters.TerminatedField(construct.Int32ub)
    )

    assert test_struct(a=1).build() == b'\x00\x00\x00\x01\x00'
    assert test_struct.parse(b'\x00\x00\x00\x01\x00').container.a == 1

    test_struct = struct.XStruct(
        'a' / adapters.TerminatedField(
            construct.Int32ub, length=4, pattern=b'\xff'
        )
    )

    assert test_struct(a=1).build() == b'\x00\x00\x00\x01\xff\xff\xff\xff'
    assert test_struct.parse(b'\x00\x00\x00\x01\xff\xff\xff\xff').container.a == 1


def test_sgstring():
    test_struct = struct.XStruct(
        'a' / adapters.SGString()
    )

    assert test_struct(a='test').build() == b'\x00\x04test\x00'
    assert test_struct.parse(b'\x00\x04test\x00').container.a == 'test'


def test_fieldin():
    test_struct = struct.XStruct(
        'a' / construct.Int32ub,
        'b' / construct.IfThenElse(
            adapters.FieldIn('a', [1, 2, 3]),
            construct.Int32ub, construct.Int16ub
        )
    )

    assert test_struct(a=1, b=2).build() == b'\x00\x00\x00\x01\x00\x00\x00\x02'
    assert test_struct(a=2, b=2).build() == b'\x00\x00\x00\x02\x00\x00\x00\x02'
    assert test_struct(a=3, b=2).build() == b'\x00\x00\x00\x03\x00\x00\x00\x02'
    assert test_struct(a=4, b=2).build() == b'\x00\x00\x00\x04\x00\x02'



"""
Cryptography portion used for Title Channel aka Auxiliary Stream
"""
import hmac
import hashlib
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


class AuxiliaryStreamCrypto(object):
    _backend = default_backend()

    def __init__(self, crypto_key, hash_key, server_iv, client_iv):
        """
        Initialize Auxiliary Stream Crypto-context.
        """

        self._encrypt_key = crypto_key
        self._hash_key = hash_key
        self._server_iv = server_iv
        self._client_iv = client_iv

        self._server_cipher = Cipher(
            algorithms.AES(self._encrypt_key),
            modes.CBC(self._server_iv),
            backend=AuxiliaryStreamCrypto._backend
        )

        self._server_encryptor = self._server_cipher.encryptor()
        self._server_decryptor = self._server_cipher.decryptor()

        self._client_cipher = Cipher(
            algorithms.AES(self._encrypt_key),
            modes.CBC(self._client_iv),
            backend=AuxiliaryStreamCrypto._backend
        )

        self._client_encryptor = self._client_cipher.encryptor()
        self._client_decryptor = self._client_cipher.decryptor()

    @classmethod
    def from_connection_info(cls, connection_info):
        """
        Initialize Crypto context via AuxiliaryStream-message
        connection info.
        """
        return cls(
            connection_info.crypto_key,
            connection_info.sign_hash,
            connection_info.server_iv,
            connection_info.client_iv
        )

    def encrypt(self, plaintext):
        """
        Encrypts plaintext with AES-128-CBC

        No padding is added here, data has to be aligned to
        block size (16 bytes).

        Args:
            plaintext (bytes): The plaintext to encrypt.

        Returns:
            bytes: Encrypted Data
        """
        return AuxiliaryStreamCrypto._crypt(self._client_encryptor, plaintext)

    def encrypt_server(self, plaintext):
        return AuxiliaryStreamCrypto._crypt(self._server_encryptor, plaintext)

    def decrypt(self, ciphertext):
        """
        Decrypts ciphertext

        No padding is removed here.

        Args:
            ciphertext (bytes): Ciphertext to be decrypted

        Returns:
            bytes: Decrypted data
        """
        return AuxiliaryStreamCrypto._crypt(self._server_decryptor, ciphertext)

    def decrypt_client(self, ciphertext):
        return AuxiliaryStreamCrypto._crypt(self._client_decryptor, ciphertext)

    def hash(self, data):
        """
        Securely hashes data with HMAC SHA-256

        Args:
            data (bytes): The data to securely hash.

        Returns:
            bytes: Hashed data
        """
        return AuxiliaryStreamCrypto._secure_hash(self._hash_key, data)

    def verify(self, data, secure_hash):
        """
        Verifies that the given data generates the given secure_hash

        Args:
            data (bytes): The data to validate.
            secure_hash (bytes): The secure hash to validate against.

        Returns:
            bool: True on success, False otherwise
        """
        return secure_hash == self.hash(data)

    @staticmethod
    def _secure_hash(key, data):
        return hmac.new(key, data, hashlib.sha256).digest()

    @staticmethod
    def _crypt(cryptor, data):
        return cryptor.update(data)



"""
Managers for handling different ServiceChannels.

If a manager for a specific :class:`ServiceChannel` is attached,
incoming messages get forwarded there, otherways they are discarded.

Managers can be attached by calling `add_manager()` on the :class:`Console` object (see example)
Methods of manager are available through console-context

Example:
    How to add a manager::

        discovered = Console.discover(timeout=1)
        if len(discovered):
            console = discovered[0]

            # Add manager, optionally passing initialization parameter
            some_arg_for_manager_init = 'example'
            console.add_manager(MediaManager, additional_arg=some_arg_for_manager_init)

            console.connect()
            if console.connection_state != ConnectionState.Connected:
                print("Connection failed")
                sys.exit(1)
            console.wait(1)

            # Call manager method
            console.media_command(0x54321, MediaControlCommand.PlayPauseToggle, 0)

            try:
                console.protocol.serve_forever()
            except KeyboardInterrupt:
                pass
        else:
            print("No consoles discovered")
            sys.exit(1)
"""
import time
import logging
from xbox.sg import factory
from xbox.sg.enum import MessageType, ServiceChannel, AckStatus, TextResult
from xbox.sg.utils.events import Event

log = logging.getLogger(__name__)


class Manager(object):
    __namespace__ = ''

    def __init__(self, console, channel):
        """
        Don't use directly!
        INTERNALLY called by the parent :class:`Console`!

        Args:
            console (:class:`Console`): Console object, internally passed
                by `Console.add_manager`
            channel (:class:`ServiceChannel`): Service channel
        """
        self.console = console
        self.console.protocol.on_message += self._pre_on_message
        self.console.protocol.on_json += self._pre_on_json
        self._channel = channel

    def _pre_on_message(self, msg, channel):
        if channel == self._channel:
            self._on_message(msg, channel)

    def _pre_on_json(self, data, channel):
        if channel == self._channel:
            self._on_json(data, channel)

    def _on_message(self, msg, channel):
        """
        Managers must implement this
        """
        pass

    def _on_json(self, data, channel):
        """
        Managers must implement this
        """
        pass

    def _send_message(self, msg):
        """
        Internal method to send messages to initialized Service Channel

        Args:
            msg (:class:`XStructObj`): Message
        """
        return self.console.protocol.send_message(msg, channel=self._channel)

    def _send_json(self, data):
        """
        Internal method to send JSON messages to initialized Service Channel

        Args:
            data (dict): JSON message
        """
        return self.console.protocol.json(data, channel=self._channel)


class InputManagerError(Exception):
    """
    Exception thrown by InputManager
    """
    pass


class InputManager(Manager):
    __namespace__ = 'input'

    def __init__(self, console):
        """
        Input Manager (ServiceChannel.SystemInput)

        Args:
             console (:class:`.Console`): Console object, internally passed
                                          by `Console.add_manager`

        """
        super(InputManager, self).__init__(console, ServiceChannel.SystemInput)

    def _on_message(self, msg, channel):
        """
        Internal handler method to receive messages from SystemInput Channel

        Args:
            msg (:class:`XStructObj`): Message
            channel (:class:`ServiceChannel`): Service channel
        """
        raise InputManagerError("Unexpected message received on InputManager")

    def gamepad_input(self, buttons, l_trigger=0, r_trigger=0, l_thumb_x=0,
                      l_thumb_y=0, r_thumb_x=0, r_thumb_y=0):
        """
        Send gamepad input

        Args:
            buttons (:class:`GamePadButton`): Gamepad buttons bits
            l_trigger (int): Left trigger value
            r_trigger (int): Right trigger value
            l_thumb_x (int): Left thumbstick X-axis value
            l_thumb_y (int): Left thumbstick Y-axis value
            r_thumb_x (int): Right thumbstick X-axis value
            r_thumb_y (int): Right thumbstick Y-axis value

        Returns: None
        """
        ts = int(time.time())
        msg = factory.gamepad(
            ts, buttons, l_trigger, r_trigger, l_thumb_x, l_thumb_y,
            r_thumb_x, r_thumb_y
        )
        return self._send_message(msg)


class MediaManagerError(Exception):
    """
    Exception thrown by MediaManager
    """
    pass


class MediaManager(Manager):
    __namespace__ = 'media'

    def __init__(self, console):
        """
        Media Manager (ServiceChannel.SystemMedia)

        Args:
             console (:class:`.Console`): Console object, internally passed
                                          by `Console.add_manager`

        """
        super(MediaManager, self).__init__(console, ServiceChannel.SystemMedia)
        self._media_state = None
        self.on_media_state = Event()
        self.on_media_command_result = Event()
        self.on_media_controller_removed = Event()

    def _on_message(self, msg, channel):
        """
        Internal handler method to receive messages from SystemMedia Channel

        Args:
            msg (:class:`XStructObj`): Message
            channel (:class:`ServiceChannel`): Service channel
        """
        msg_type = msg.header.flags.msg_type
        payload = msg.protected_payload
        if msg_type == MessageType.MediaState:
            log.debug('Received MediaState message')
            self._media_state = payload
            self.on_media_state(self.media_state)

        elif msg_type == MessageType.MediaCommandResult:
            log.debug('Received MediaCommandResult message')
            self.on_media_command_result(payload)

        elif msg_type == MessageType.MediaControllerRemoved:
            title_id = payload.title_id
            log.debug('Received MediaControllerRemoved message, title id: 0x%x', title_id)
            if self.title_id == title_id:
                log.debug('Clearing MediaState')
                self._media_state = None
            self.on_media_controller_removed(payload)

        else:
            raise MediaManagerError("Unexpected message received on MediaManager")

    @property
    def media_state(self):
        """
        Media state payload

        Returns:
            :class:`Container`: Media state payload
        """
        return self._media_state

    @property
    def active_media(self):
        """
        Check whether console has active media

        Returns:
            bool: `True` if media is active, `False` if not
        """
        return self.media_state is not None

    @property
    def title_id(self):
        """
        Title Id of active media

        Returns:
            int: Title Id
        """
        if self.media_state:
            return self.media_state.title_id

    @property
    def aum_id(self):
        """
        Application user model Id of active media

        Returns:
            str: Aum Id
        """
        if self.media_state:
            return self.media_state.aum_id

    @property
    def asset_id(self):
        """
        Asset Id of active media

        Returns:
            str: Asset Id
        """
        if self.media_state:
            return self.media_state.asset_id

    @property
    def media_type(self):
        """
        Media type of active media

        Returns:
            :class:`MediaType`: Media type
        """
        if self.media_state:
            return self.media_state.media_type

    @property
    def sound_level(self):
        """
        Sound level of active media

        Returns:
            :class:`SoundLevel`: Sound level
        """
        if self.media_state:
            return self.media_state.sound_level

    @property
    def enabled_commands(self):
        """
        Enabled MediaCommands bitmask

        Returns:
            :class:`MediaControlCommand`: Bitmask of enabled commands
        """
        if self.media_state:
            return self.media_state.enabled_commands

    @property
    def playback_status(self):
        """
        Title Id of active media

        Returns:
            :class:`MediaPlaybackStatus`: Playback status
        """
        if self.media_state:
            return self.media_state.playback_status

    @property
    def rate(self):
        """
        Playback rate of active media

        Returns:
            float: Playback rate
        """
        if self.media_state:
            return self.media_state.rate

    @property
    def position(self):
        """
        Playback position of active media

        Returns:
            int: Playback position in microseconds
        """
        if self.media_state:
            return self.media_state.position

    @property
    def media_start(self):
        """
        Media start position of active media

        Returns:
            int: Media start position in microseconds
        """
        if self.media_state:
            return self.media_state.media_start

    @property
    def media_end(self):
        """
        Media end position of active media

        Returns:
            int: Media end position in microseconds
        """
        if self.media_state:
            return self.media_state.media_end

    @property
    def min_seek(self):
        """
        Minimum seek position of active media

        Returns:
            int: Minimum position in microseconds
        """
        if self.media_state:
            return self.media_state.min_seek

    @property
    def max_seek(self):
        """
        Maximum seek position of active media

        Returns:
            int: Maximum position in microseconds
        """
        if self.media_state:
            return self.media_state.max_seek

    @property
    def metadata(self):
        """
        Media metadata of active media

        Returns:
            :class:`Container`: Media metadata
        """
        if self.media_state:
            return self.media_state.metadata

    def media_command(self, title_id, command, request_id=0,
                      seek_position=None):
        """
        Send media command

        Args:
            title_id (int): Title Id
            command (:class:`MediaControlCommand`): Media Command
            request_id (int): Incrementing Request Id
            seek_position (int): (optional) Seek position

        Returns: None
        """
        msg = factory.media_command(
            request_id, title_id, command, seek_position
        )
        return self._send_message(msg)


class TextManagerError(Exception):
    """
    Exception thrown by TextManager
    """
    pass


class TextManager(Manager):
    __namespace__ = 'text'

    def __init__(self, console):
        """
        Text Manager (ServiceChannel.SystemText)

        Args:
             console (:class:`.Console`): Console object, internally passed
                                          by `Console.add_manager`

        """
        super(TextManager, self).__init__(console, ServiceChannel.SystemText)

        self.session_config = None
        self.current_session_input = None
        self.last_session_ack = None
        self._current_text_version = 0

        self.on_systemtext_configuration = Event()
        self.on_systemtext_input = Event()
        self.on_systemtext_done = Event()

    def _on_message(self, msg, channel):
        """
        Internal handler method to receive messages from SystemText Channel

        Args:
            msg (:class:`XStructObj`): Message
            channel (:class:`ServiceChannel`): Service channel
        """
        msg_type = msg.header.flags.msg_type
        payload = msg.protected_payload
        session_id = payload.text_session_id

        if msg_type == MessageType.SystemTextConfiguration:
            self.reset_session()
            self.session_config = payload
            self.on_systemtext_configuration(payload)

        elif msg_type == MessageType.SystemTextInput:
            # Assign console input msg
            self.current_session_input = payload
            self.current_text_version = payload.submitted_version
            self.send_systemtext_ack(self.text_session_id,
                                     self.current_text_version)
            self.on_systemtext_input(payload)

        elif msg_type == MessageType.SystemTextAck:
            self.current_text_version = payload.text_version_ack

        elif msg_type == MessageType.SystemTextDone:
            if session_id == self.text_session_id:
                self.reset_session()
            elif session_id == 0:
                # SystemTextDone for session 0 is sent by console
                # No clue what it means, if anything
                pass
            else:
                pass
                # log.debug('Received DONE msg for inactive session %i' % session_id)

            self.on_systemtext_done(payload)

        elif msg_type in [MessageType.TitleTextConfiguration,
                          MessageType.TitleTextInput,
                          MessageType.TitleTextSelection]:
            raise TextManagerError('Received TitleTextConfiguration, unhandled')
        else:
            raise TextManagerError(
                "Unexpected message received on TextManager"
            )

    @property
    def got_active_session(self):
        """
        Check whether a text session is active

        Returns:
            bool: Returns `True` if any text session is active, `False` otherwise
        """
        return self.session_config is not None

    @property
    def current_text_version(self):
        """
        Current Text version

        Returns:
             int: Current Text Version
        """
        return self._current_text_version

    @current_text_version.setter
    def current_text_version(self, value):
        if value > self.current_text_version:
            self._current_text_version = value

    @property
    def text_session_id(self):
        """
        Current Text session id

        Returns:
            int: Text session id if existing, `None` otherwise
        """
        if self.session_config:
            return self.session_config.text_session_id

    @property
    def text_options(self):
        """
        Current Text options

        Returns:
            :class:`TextOption`: Text options if existing, `None` otherwise
        """
        if self.session_config:
            return self.session_config.text_options

    @property
    def text_input_scope(self):
        """
        Current Text input scope

        Returns:
            :class:`TextInputScope`: Text input scope if existing, `None` otherwise
        """
        if self.session_config:
            return self.session_config.input_scope

    @property
    def max_text_length(self):
        """
        Maximum Text length

        Returns:
            int: Max text length if existing, `None` otherwise
        """
        if self.session_config:
            return self.session_config.max_text_length

    @property
    def text_locale(self):
        """
        Test

        Returns:
            str: Text locale if existing, `None` otherwise
        """
        if self.session_config:
            return self.session_config.locale

    @property
    def text_prompt(self):
        """
        Test

        Returns:
            str: Text prompt if existing, `None` otherwise
        """
        if self.session_config:
            return self.session_config.prompt

    def reset_session(self):
        """
        Delete cached text-session config, -input and -ack messages

        Returns:
            None
        """
        self.session_config = None
        self.current_session_input = None
        self.last_session_ack = None
        self.current_text_version = 0

    def finish_text_input(self):
        """
        Finishes current text session.

        Returns:
            None
        """
        self.send_systemtext_done(
            session_id=self.text_session_id,
            version=self.current_session_input.submitted_version,
            flags=0,
            result=TextResult.Accept
        )

    def send_systemtext_input(self, text):
        """
        Sends text input

        Args:
            text (str): Text string to send

        Raises:
            TextManagerError: If message was not
            acknowledged via AckMsg or SystemTextAck

        Returns:
            int: Member of :class:`AckStatus`
        """
        new_version = self.current_text_version + 1
        msg = factory.systemtext_input(
            session_id=self.text_session_id,
            base_version=self.current_text_version,
            submitted_version=new_version,
            total_text_len=len(text),
            selection_start=-1,
            selection_length=-1,
            flags=0,
            text_chunk_byte_start=0,
            text_chunk=text,
            delta=None
        )

        ack_status = self._send_message(msg)
        if ack_status != AckStatus.Processed:
            raise TextManagerError('InputMsg was not acknowledged: %s' % msg)

        # Assign client system input msg
        self.current_session_input = msg.protected_payload
        return ack_status

    def send_systemtext_ack(self, session_id, version):
        """
        Acknowledges a SystemText message sent from the console

        Args:
            session_id (int): Current text session id
            version (int): Text version to ack

        Returns:
            int: Member of :class:`AckStatus`
        """
        msg = factory.systemtext_ack(session_id, version)
        return self._send_message(msg)

    def send_systemtext_done(self, session_id, version, flags, result):
        """
        Informs the console that a text session is done.

        Result field tells wether text input should be
        accepted or cancelled.

        Args:
            session_id (int): Current text session id
            version (id): Last acknowledged text version
            flags (int): Flags
            result (:class:`TextResult`): Member to :class:`TextResult`

        Returns:
            int: Member of :class:`AckStatus`
        """
        msg = factory.systemtext_done(session_id, version, flags, result)
        ack_status = self._send_message(msg)
        if ack_status != AckStatus.Processed:
            raise TextManagerError('DoneMsg was not acknowledged: %s' % msg)

        return ack_status



"""
StumpManager - Handling TV Streaming / IR control commands
"""
import random
import logging
import requests

from xbox.sg.utils.events import Event
from xbox.sg.enum import ServiceChannel
from xbox.sg.manager import Manager

from xbox.stump.enum import Message, Notification, Source, SourceHttpQuery, Quality
from xbox.stump import json_model


log = logging.getLogger(__name__)


class StumpException(Exception):
    """
    Exception thrown by StumpManager
    """
    pass


class StumpManager(Manager):
    __namespace__ = 'stump'

    def __init__(self, console):
        """
        Title Manager (ServiceChannel.SystemInputTVRemote)

        Args:
             console (:class:`.Console`): Console object, internally passed
                                          by `Console.add_manager`
        """
        super(StumpManager, self).__init__(
            console, ServiceChannel.SystemInputTVRemote
        )

        self._msg_id_idx = 0
        self._msg_id_prefix = random.randint(0, 0x7FFFFFFF)

        self._appchannel_lineups = None
        self._appchannel_data = None
        self._appchannel_program_data = None
        self._stump_config = None
        self._stump_headend_info = None
        self._stump_livetv_info = None
        self._stump_program_info = None
        self._stump_recent_channels = None
        self._stump_tuner_lineups = None

        self.on_response = Event()
        self.on_notification = Event()
        self.on_error = Event()

    @property
    def msg_id(self):
        """
        Internal method for generating message IDs.

        Returns:
            str: A new message ID.
        """
        self._msg_id_idx += 1
        return '{:08x}.{:d}'.format(self._msg_id_prefix, self._msg_id_idx)

    def generate_stream_url(self, source, xuid=None, quality=Quality.BEST):
        """
        Generate TV Streaming URL (Dvb-USB tuner, not hdmi-in)

        Args:
            source (:class:`Source`): Streaming source
            xuid (str): optional, XUID
            quality (:class:`Quality`): Streaming quality

        Returns:
             str: HTTP URL
        """
        src_map = {
            Source.HDMI: SourceHttpQuery.HDMI,
            Source.TUNER: SourceHttpQuery.TUNER
        }

        if isinstance(quality, str):
            quality = Quality(quality)
        if isinstance(source, str):
            source = Source(source)

        if not self.streaming_port:
            raise StumpException('No streaming port available to generate URL')
        elif source not in (Source.TUNER, Source.HDMI):
            raise StumpException('Invalid source parameter to generate URL')

        url = 'http://{address}:{port}/{url_src}'.format(
            address=self.console.address,
            port=self.streaming_port,
            url_src=src_map.get(source).value
        )

        params = {
            'xuid': xuid,
            'quality': quality.value
        }

        r = requests.Request(method='GET', url=url, params=params).prepare()
        return r.url

    def _on_json(self, data, channel):
        """
        Internal handler for JSON messages received by the core protocol.

        Args:
            data (dict): The JSON object that was received.
            channel (int): The channel this message was received on.

        Returns:
            None.
        """
        msg = json_model.deserialize_stump_message(data)

        if msg.msgid:
            self.console.protocol._set_result(msg.msgid, msg)

        if isinstance(msg, json_model.StumpError):
            log.debug("Error msg: {}".format(msg))
            self._on_error(msg)

        elif isinstance(msg, json_model.StumpResponse):
            log.debug("Response msg: {}".format(msg))
            self._on_response(msg.response, msg)

        elif isinstance(msg, json_model.StumpNotification):
            log.debug("Notification msg: {}".format(msg))
            self._on_notification(msg.notification, msg)

        else:
            log.warning("Unknown stump message: {}".format(data))

    def _on_response(self, message_type, data):
        """
        Internal response handler. For logging purposes.

        Args:
            message_type (str/Message): The message type.
            data (`StumpResponse`): The raw message.
        """
        if isinstance(message_type, str):
            message_type = Message(message_type)

        params = data.params

        if Message.CONFIGURATION == message_type:
            self._stump_config = params
        elif Message.HEADEND_INFO == message_type:
            self._stump_headend_info = params
        elif Message.LIVETV_INFO == message_type:
            self._stump_livetv_info = params
        elif Message.TUNER_LINEUPS == message_type:
            self._stump_tuner_lineups = params
        elif Message.RECENT_CHANNELS:
            self._stump_recent_channels = params
        elif Message.PROGRAMM_INFO:
            self._stump_program_info = params
        elif Message.APPCHANNEL_LINEUPS:
            self._appchannel_lineups = params
        elif Message.APPCHANNEL_DATA:
            self._appchannel_data = params
        elif Message.APPCHANNEL_PROGRAM_DATA:
            self._appchannel_program_data = params
        elif Message.SEND_KEY or \
                Message.ENSURE_STREAMING_STARTED or \
                Message.SET_CHANNEL:
            # Refer to returned data from request_* methods
            pass
        log.info("Received {} response".format(message_type))
        self.on_response(message_type, data)

    def _on_notification(self, notification, data):
        """
        Internal notification handler. For logging purposes.

        Args:
            notification (str/Notification): The notification type.
            data (`StumpNotification`): The raw message.
        """
        if isinstance(notification, str):
            notification = Notification(notification)

        log.info("Received {} notification: {}".format(notification, data))
        self.on_notification(notification, data)

    def _on_error(self, data):
        """
        Internal error handler.

        Args:
            data (`StumpError`): The error dictionary from the Message.

        Returns:
            None.
        """
        log.error("Error: {}".format(data))
        self.on_error(data)

    def _send_stump_message(self, name, params=None, msgid=None, timeout=3):
        """
        Internal method for sending JSON messages over the core protocol.

        Handles message IDs as well as waiting for results.

        Args:
            name (Enum): Request name
            params (dict): The message parameters to send.
            msgid (str): Message identifier
            timeout (int): Timeout in seconds

        Returns:
            dict: The received result.
        """
        if not msgid:
            msgid = self.msg_id

        msg = dict(msgid=msgid, request=name.value, params=params)
        self._send_json(msg)

        result = self.console.protocol._await_ack(msgid, timeout)
        if not result:
            raise StumpException("Message \'{}\': \'{}\' got no response!".format(msgid, name))

        return result

    def request_stump_configuration(self):
        """
        Request device configuration from console.

        The configuration holds info about configured, by Xbox controlable \
        devices (TV, AV, STB).

        Returns:
            dict: The received result.
        """
        return self._send_stump_message(Message.CONFIGURATION)

    def request_headend_info(self):
        """
        Request available headend information from console.

        Returns:
            dict: The received result.
        """
        return self._send_stump_message(Message.HEADEND_INFO)

    def request_live_tv_info(self):
        """
        Request LiveTV information from console.

        Holds information about currently tuned channel, streaming-port etc.

        Returns:
            dict: The received result.
        """
        return self._send_stump_message(Message.LIVETV_INFO)

    def request_program_info(self):
        """
        Request program information.

        NOTE: Not working?!

        Returns:
            dict: The received result.
        """
        return self._send_stump_message(Message.PROGRAMM_INFO)

    def request_tuner_lineups(self):
        """
        Request Tuner Lineups from console.

        Tuner lineups hold information about scanned / found channels.

        Returns:
            dict: The received result.
        """
        return self._send_stump_message(Message.TUNER_LINEUPS)

    def request_app_channel_lineups(self):
        """
        Request AppChannel Lineups.

        Returns:
            dict: The received result.
        """
        return self._send_stump_message(Message.APPCHANNEL_LINEUPS)

    def request_app_channel_data(self, provider_id, channel_id):
        """
        Request AppChannel Data.

        Args:
            provider_id (str): Provider ID.
            channel_id (str): Channel ID.

        Returns:
            dict: The received result.
        """
        return self._send_stump_message(
            Message.APPCHANNEL_DATA,
            params={
                'providerId': provider_id,
                'channelId': channel_id,
                'id': channel_id
            }
        )

    def request_app_channel_program_data(self, provider_id, program_id):
        """
        Request AppChannel Program Data.

        Args:
            provider_id (str): Provider ID.
            program_id (str): Program Id.

        Returns:
            dict: The received result.
        """
        return self._send_stump_message(
            Message.APPCHANNEL_PROGRAM_DATA,
            params={
                'providerId': provider_id,
                'programId': program_id
            }
        )

    def set_stump_channel_by_id(self, channel_id, lineup_id):
        """
        Switch to channel by providing channel ID and lineup ID.

        Args:
            channel_id (str): Channel ID.
            lineup_id (str): Lineup ID.

        Returns:
            dict: The received result.
        """
        return self._send_stump_message(
            Message.SET_CHANNEL,
            params={
                'channelId': channel_id,
                'lineupInstanceId': lineup_id
            }
        )

    def set_stump_channel_by_name(self, channel_name):
        """
        Switch to channel by providing channel name.

        Args:
            channel_name (str): Channel name to switch to.

        Returns:
            dict: The received result.
        """
        return self._send_stump_message(
            Message.SET_CHANNEL,
            params={
                'channel_name': channel_name
            }
        )

    def request_recent_channels(self, first, count):
        """
        Request a list of recently watched channels.

        Args:
            first (int): Where to start enumeration.
            count (int): Number of channels to request.

        Returns:
            dict: The received result.
        """
        return self._send_stump_message(
            Message.RECENT_CHANNELS,
            params={
                'startindex': first,
                'count': count
            }
        )

    def send_stump_key(self, button, device_id=None, **kwargs):
        """
        Send a remote control button to configured device via \
        Xbox's IR Blaster / Kinect / whatev.

        Args:
            button (str): Button to send.
            device_id (str, int): Device ID of device to control.

        Returns:
            dict: The received result.
        """
        params = dict(button_id=button)

        if device_id:
            params['device_id'] = device_id

        if kwargs:
            params.update(**kwargs)

        return self._send_stump_message(
            Message.SEND_KEY,
            params=params
        )

    def request_ensure_stump_streaming_started(self, source):
        """
        Ensure that streaming started on desired tuner type

        Args:
            source (str): Tuner Source to check for.
                          Member of :class:`Source`

        Returns:
            dict: The received result.
        """
        return self._send_stump_message(
            Message.ENSURE_STREAMING_STARTED,
            params={
                'source': source
            }
        )

    @property
    def headend_locale(self):
        if self._stump_headend_info:
            return self._stump_headend_info.headendLocale

    @property
    def streaming_port(self):
        if self._stump_livetv_info:
            return self._stump_livetv_info.streamingPort

    @property
    def is_hdmi_mode(self):
        if self._stump_livetv_info:
            return self._stump_livetv_info.inHdmiMode

    @property
    def current_tunerchannel_type(self):
        if self._stump_livetv_info:
            return self._stump_livetv_info.tunerChannelType



import sys

from gevent import monkey
monkey.patch_all()

from xbox.scripts import main_cli  # noqa: E402


def main():
    print('Starting REST server from dedicated script')
    return sys.exit(main_cli.main('rest'))


if __name__ == '__main__':
    main()



"""
Text smartglass client
"""
import sys
import logging
import gevent.socket

LOGGER = logging.getLogger(__name__)


def on_text_config(payload):
    pass


def on_text_input(console, payload):
    print(
        "\n\nWAITING FOR TEXT INPUT - "
        "SHELL WILL KEEP SCROLLING\n"
        "Prompt: %s\n" % console.text.text_prompt
    )
    gevent.socket.wait_read(sys.stdin.fileno())
    text = input()
    console.send_systemtext_input(text)
    console.finish_text_input()


def on_text_done(payload):
    pass



import pytest
from binascii import unhexlify
from xbox.sg.enum import PublicKeyType


def test_generate_random_iv(crypto):
    rand_iv = crypto.generate_iv()
    rand_iv_2 = crypto.generate_iv()

    assert len(rand_iv) == 16
    assert len(rand_iv_2) == 16
    assert rand_iv != rand_iv_2


def test_generate_seeded_iv(crypto):
    seed = unhexlify('000102030405060708090A0B0C0D0E0F')
    seed2 = unhexlify('000F0E0D0C0B0A090807060504030201')

    seed_iv = crypto.generate_iv(seed)
    seed_iv_dup = crypto.generate_iv(seed)
    seed_iv_2 = crypto.generate_iv(seed2)

    assert len(seed_iv) == 16
    assert seed_iv == seed_iv_dup
    assert seed_iv != seed_iv_2


def test_encrypt_decrypt(crypto):
    plaintext = b'Test String\x00\x00\x00\x00\x00'
    seed = unhexlify('000102030405060708090A0B0C0D0E0F')
    seed_iv = crypto.generate_iv(seed)
    encrypt = crypto.encrypt(seed_iv, plaintext)
    decrypt = crypto.decrypt(seed_iv, encrypt)

    assert plaintext == decrypt
    assert plaintext != encrypt


def test_hash(crypto):
    plaintext = b'Test String\x00\x00\x00\x00\x00'
    seed = unhexlify('000102030405060708090A0B0C0D0E0F')
    seed_iv = crypto.generate_iv(seed)
    encrypt = crypto.encrypt(seed_iv, plaintext)
    hash = crypto.hash(encrypt)
    hash_dup = crypto.hash(encrypt)
    verify = crypto.verify(encrypt, hash)

    assert hash == hash_dup
    assert verify is True


def test_from_bytes(public_key_bytes, public_key):
    from xbox.sg.crypto import Crypto
    c1 = Crypto.from_bytes(public_key_bytes)
    c2 = Crypto.from_bytes(public_key_bytes, PublicKeyType.EC_DH_P256)

    # invalid public key type passed
    with pytest.raises(ValueError):
        Crypto.from_bytes(public_key_bytes, PublicKeyType.EC_DH_P521)
    # invalid keylength
    with pytest.raises(ValueError):
        Crypto.from_bytes(public_key_bytes[5:])
    # invalid parameter
    with pytest.raises(ValueError):
        Crypto.from_bytes(123)

    assert c1.foreign_pubkey.public_numbers() == public_key.public_numbers()
    assert c2.foreign_pubkey.public_numbers() == public_key.public_numbers()


def test_from_shared_secret(shared_secret_bytes):
    from xbox.sg.crypto import Crypto
    c = Crypto.from_shared_secret(shared_secret_bytes)

    # invalid length
    with pytest.raises(ValueError):
        c.from_shared_secret(shared_secret_bytes[1:])

    # invalid parameter
    with pytest.raises(ValueError):
        c.from_shared_secret(123)

    assert c._encrypt_key == unhexlify(b'82bba514e6d19521114940bd65121af2')
    assert c._iv_key == unhexlify(b'34c53654a8e67add7710b3725db44f77')
    assert c._hash_key == unhexlify(
        b'30ed8e3da7015a09fe0f08e9bef3853c0506327eb77c9951769d923d863a2f5e'
    )



from xbox.scripts import pcap


def test_pcap_filter(pcap_filepath):
    packets = list(pcap.packet_filter(pcap_filepath))

    assert len(packets) == 26


def test_run_parser(pcap_filepath, crypto):
    pcap.parse(pcap_filepath, crypto)



"""
Custom construct fields and utilities
"""
import construct


COMPILED = True


class XStruct(construct.Subconstruct):
    def __init__(self, *args, **kwargs):
        struct = construct.Struct(*args, **kwargs)
        super(XStruct, self).__init__(struct)
        self.compiled = self.compile() if COMPILED else None

    def parse(self, data, **contextkw):
        res = super(XStruct, self).parse(data, **contextkw)
        return XStructObj(self, res)

    def _parse(self, stream, context, path):
        if self.compiled:
            res = self.compiled._parse(stream, context, path)
        else:
            res = self.subcon._parse(stream, context, path)

        return res

    def _emitparse(self, code):
        return self.subcon._emitparse(code)

    def _find(self, item):
        sub = next(
            (s for s in self.subcon.subcons if s.name == item), None
        )
        if sub:
            return sub

        return None

    def __call__(self, **kwargs):
        return XStructObj(self)(**kwargs)

    def __contains__(self, item):
        return self._find(item) is not None

    def __getattr__(self, item):
        subcon = self._find(item)

        if subcon:
            return subcon

        return super(XStruct, self).__getattribute__(item)


class XStructObj(construct.Subconstruct):
    __slots__ = ['_obj']

    def __init__(self, struct, container=None):
        super(XStructObj, self).__init__(struct)
        self._obj = container if container else construct.Container({
            sc.name: None for sc in self.subcon.subcon.subcons
        })

    @property
    def container(self):
        return self._obj

    def __call__(self, **kwargs):
        self._obj.update(kwargs)
        return self

    def _parse(self, stream, context, path):
        self._obj = self.subcon._parse(stream, context, path)
        return self

    def build(self, **contextkw):
        return self.subcon.build(self._obj, **contextkw)

    def _build(self, stream, context, path):
        return self.subcon._build(self._obj, stream, context, path)

    def __getattr__(self, item):
        if item in self._obj:
            return getattr(self._obj, item)
        return super(XStructObj, self).__getattribute__(item)

    def __repr__(self):
        return '<XStructObj: %s>(%s)' % (
            self.subcon.name, self._obj
        )


def flatten(container):
    """
    Flattens `StructWrap` objects into just `Container`'s.

    Recursively walks down each value of the `Container`, flattening
    possible `StructWrap` objects.

    Args:
        container (Container): The container to flatten.

    Returns:
        `Container`: A flattened container.
    """
    obj = container.copy()
    for k, v in obj.items():
        if isinstance(v, XStructObj):
            obj[k] = flatten(v.container)

    return obj



"""
Wrapper around gevent's greenlets
"""
import gevent


class Event(object):
    def __init__(self, green=False):
        self.handlers = []
        self.green = green

    def add(self, handler):
        if not callable(handler):
            raise TypeError("Handler should be callable")
        self.handlers.append(handler)

    def remove(self, handler):
        if handler in self.handlers:
            self.handlers.remove(handler)

    def __iadd__(self, handler):
        self.add(handler)
        return self

    def __isub__(self, handler):
        self.remove(handler)
        return self

    def __call__(self, *args, **kwargs):
        for handler in self.handlers:
            if self.green:
                gevent.spawn(handler, *args, **kwargs)
            else:
                handler(*args, **kwargs)



from flask import current_app as app
from flask import request
from http import HTTPStatus
from xbox.sg import enum
from ..decorators import console_exists, console_connected
from ..consolewrap import ConsoleWrap
from . import routes


@routes.route('/device')
def device_overview():
    addr = request.args.get('addr')
    discovered = ConsoleWrap.discover(addr=addr).copy()

    liveids = [d.liveid for d in discovered]
    for i, c in enumerate(app.console_cache.values()):
        if c.liveid in liveids:
            # Refresh existing entries
            index = liveids.index(c.liveid)

            if c.device_status != discovered[index].device_status:
                app.console_cache[c.liveid] = ConsoleWrap(discovered[index])
            del discovered[index]
            del liveids[index]
        elif c.liveid not in liveids:
            # Set unresponsive consoles to Unavailable
            app.console_cache[c.liveid].console.device_status = enum.DeviceStatus.Unavailable

    # Extend by new entries
    for d in discovered:
        app.console_cache.update({d.liveid: ConsoleWrap(d)})

    # Filter for specific console when ip address query is supplied (if available)
    data = {console.liveid: console.status for console in app.console_cache.values()
            if (addr and console.status.get('address') == addr) or not addr}
    return app.success(devices=data)


@routes.route('/device/<liveid>/poweron')
def poweron(liveid):
    addr = request.args.get('addr')
    ConsoleWrap.power_on(liveid, addr=addr)
    return app.success()


"""
Require enumerated console
"""


@routes.route('/device/<liveid>')
@console_exists
def device_info(console):
    return app.success(device=console.status)


@routes.route('/device/<liveid>/connect')
@console_exists
def force_connect(console):
    try:
        userhash = ''
        xtoken = ''
        if app.authentication_mgr.authenticated and not request.args.get('anonymous'):
            userhash = app.authentication_mgr.userinfo.userhash
            xtoken = app.authentication_mgr.xsts_token.jwt
        state = console.connect(userhash, xtoken)
    except Exception as e:
        return app.error(str(e))

    if state != enum.ConnectionState.Connected:
        return app.error('Connection failed', connection_state=state.name)

    return app.success(connection_state=state.name)


"""
Require connected console
"""


@routes.route('/device/<liveid>/disconnect')
@console_connected
def disconnect(console):
    console.disconnect()
    return app.success()


@routes.route('/device/<liveid>/poweroff')
@console_connected
def poweroff(console):
    if not console.power_off():
        return app.error("Failed to power off")
    else:
        return app.success()


@routes.route('/device/<liveid>/console_status')
@console_connected
def console_status(console):
    status = console.console_status
    client = app.xbl_client
    # Update Title Info
    if client and status:
        for t in status['active_titles']:
            try:
                title_id = t['title_id']
                resp = app.title_cache.get(title_id)
                if not resp:
                    resp = client.titlehub.get_title_info(title_id, 'image').json()
                if resp['titles'][0]:
                    app.title_cache[title_id] = resp
                    t['name'] = resp['titles'][0]['name']
                    t['image'] = resp['titles'][0]['displayImage']
                    t['type'] = resp['titles'][0]['type']
            except Exception:
                pass
    return app.success(console_status=status)


@routes.route('/device/<liveid>/launch/<path:app_id>')
@console_connected
def launch_title(console, app_id):
    console.launch_title(app_id)
    return app.success(launched=app_id)


@routes.route('/device/<liveid>/media_status')
@console_connected
def media_status(console):
    return app.success(media_status=console.media_status)


@routes.route('/device/<liveid>/ir')
@console_connected
def infrared(console):
    stump_config = console.stump_config

    devices = {}
    for device_config in stump_config.params:
        button_links = {}
        for button in device_config.buttons:
            button_links[button] = {
                'url': '/device/{0}/ir/{1}/{2}'.format(console.liveid, device_config.device_id, button),
                'value': device_config.buttons[button]
            }

        devices[device_config.device_type] = {
            'device_type': device_config.device_type,
            'device_brand': device_config.device_brand,
            'device_model': device_config.device_model,
            'device_name': device_config.device_name,
            'device_id': device_config.device_id,
            'buttons': button_links
        }

    return app.success(**devices)


@routes.route('/device/<liveid>/ir/<device_id>')
@console_connected
def infrared_available_keys(console, device_id):
    stump_config = console.stump_config
    for device_config in stump_config.params:
        if device_config.device_id != device_id:
            continue

        button_links = {}
        for button in device_config.buttons:
            button_links[button] = {
                'url': '/device/{0}/ir/{1}/{2}'.format(console.liveid, device_config.device_id, button),
                'value': device_config.buttons[button]
            }

        return app.success(
            device_type=device_config.device_type,
            device_brand=device_config.device_brand,
            device_model=device_config.device_model,
            device_name=device_config.device_name,
            device_id=device_config.device_id,
            buttons=button_links
        )

    return app.error('Device Id \'{0}\' not found'.format(device_id), HTTPStatus.BAD_REQUEST)


@routes.route('/device/<liveid>/ir/<device_id>/<button>')
@console_connected
def infrared_send(console, device_id, button):
    if not console.send_stump_key(device_id, button):
        return app.error('Failed to send button')

    return app.success(sent_key=button, device_id=device_id)


@routes.route('/device/<liveid>/media')
@console_connected
def media_overview(console):
    return app.success(commands=list(console.media_commands.keys()))


@routes.route('/device/<liveid>/media/<command>')
@console_connected
def media_command(console, command):
    cmd = console.media_commands.get(command)
    if not cmd:
        return app.error('Invalid command passed, command: {0}'.format(command), HTTPStatus.BAD_REQUEST)

    console.send_media_command(cmd)
    return app.success()


@routes.route('/device/<liveid>/media/seek/<int:seek_position>')
@console_connected
def media_command_seek(console, seek_position):
    console.send_media_command(enum.MediaControlCommand.Seek, seek_position)
    return app.success()


@routes.route('/device/<liveid>/input')
@console_connected
def input_overview(console):
    return app.success(buttons=list(console.input_keys.keys()))


@routes.route('/device/<liveid>/input/<button>')
@console_connected
def input_send_button(console, button):
    btn = console.input_keys.get(button)
    if not btn:
        return app.error('Invalid button passed, button: {0}'.format(button), HTTPStatus.BAD_REQUEST)

    console.send_gamepad_button(btn)
    return app.success()


@routes.route('/device/<liveid>/stump/headend')
@console_connected
def stump_headend_info(console):
    return app.success(headend_info=console.headend_info.params.dump())


@routes.route('/device/<liveid>/stump/livetv')
@console_connected
def stump_livetv_info(console):
    return app.success(livetv_info=console.livetv_info.params.dump())


@routes.route('/device/<liveid>/stump/tuner_lineups')
@console_connected
def stump_tuner_lineups(console):
    return app.success(tuner_lineups=console.tuner_lineups.params.dump())


@routes.route('/device/<liveid>/text')
@console_connected
def text_overview(console):
    return app.success(text_session_active=console.text_active)


@routes.route('/device/<liveid>/text/<text>')
@console_connected
def text_send(console, text):
    console.send_text(text)
    return app.success()


@routes.route('/device/<liveid>/gamedvr')
@console_connected
def gamedvr_record(console):
    """
    Default to record last 60 seconds
    Adjust with start/end query parameter
    (delta time in seconds)
    """
    try:
        start_delta = request.args.get('start', -60)
        end_delta = request.args.get('end', 0)
        console.dvr_record(int(start_delta), int(end_delta))
    except Exception as e:
        return app.error('GameDVR failed, error: {0}'.format(e))

    return app.success()


@routes.route('/device/<liveid>/nano')
@console_connected
def nano_overview(console):
    return app.success(nano_status=console.nano_status)


@routes.route('/device/<liveid>/nano/start')
@console_connected
def nano_start(console):
    console.nano_start()
    return app.success()


@routes.route('/device/<liveid>/nano/stop')
@console_connected
def nano_stop(console):
    console.nano_stop()
    return app.success()



"""
Parse a pcap packet capture and show decrypted
packets in a human-readable ways.

Requires the shared secret for that smartglass-session.
"""
import os
import shutil
import string
import textwrap
import argparse
from binascii import unhexlify

import dpkt

from xbox.sg import packer
from xbox.sg.crypto import Crypto
from xbox.sg.enum import PacketType

from construct.lib import containers
containers.setGlobalPrintFullStrings(True)


def packet_filter(filepath):
    with open(filepath, 'rb') as fh:
        for ts, buf in dpkt.pcap.Reader(fh):
            eth = dpkt.ethernet.Ethernet(buf)

            # Make sure the Ethernet data contains an IP packet
            if not isinstance(eth.data, dpkt.ip.IP):
                continue

            ip = eth.data
            if not isinstance(ip.data, dpkt.udp.UDP):
                continue

            udp = ip.data
            if udp.sport != 5050 and udp.dport != 5050:
                continue

            is_client = udp.dport == 5050

            yield(udp.data, is_client, ts)


def parse(pcap_filepath, crypto):
    width = shutil.get_terminal_size().columns
    col_width = width // 2 - 3
    wrapper = textwrap.TextWrapper(col_width, replace_whitespace=False)

    for packet, is_client, ts in packet_filter(pcap_filepath):
        try:
            msg = packer.unpack(packet, crypto)
        except Exception as e:
            print("Error: {}".format(e))
            continue

        msg_type = msg.header.pkt_type
        type_str = msg_type.name

        if msg_type == PacketType.Message:
            type_str = msg.header.flags.msg_type.name

        direction = '>' if is_client else '<'
        print(' {} '.format(type_str).center(width, direction))

        lines = str(msg).split('\n')
        for line in lines:
            line = wrapper.wrap(line)
            for i in line:
                if is_client:
                    print('{0: <{1}}'.format(i, col_width), '│')
                else:
                    print(' ' * col_width, '│', '{0}'.format(i, col_width))


def main():
    parser = argparse.ArgumentParser(
        description='Parse PCAP files and show SG sessions'
    )
    parser.add_argument('file', help='Path to PCAP')
    parser.add_argument('secret', help='Expanded secret for this session.')
    args = parser.parse_args()

    secret = args.secret
    if os.path.exists(secret):
        # Assume a file containing the secret
        with open(secret, 'rb') as fh:
            secret = fh.read()

        if all(chr(c) in string.hexdigits for c in secret):
            secret = unhexlify(secret)
    else:
        secret = unhexlify(secret)

    crypto = Crypto.from_shared_secret(secret)
    parse(args.file, crypto)


if __name__ == '__main__':
    main()



import uuid
from binascii import unhexlify

from xbox.sg import enum, constants
from xbox.sg import packer
from xbox.sg import factory


def _pack(msg, crypto):
    return packer.pack(msg, crypto)


def test_message_header():
    msg = factory._message_header(
        msg_type=enum.MessageType.Ack,
        target_participant_id=2,
        source_participant_id=3,
        need_ack=False,
        is_fragment=False,
        channel_id=0xFFFF
    )
    packed = msg.build()

    assert msg.flags.msg_type == enum.MessageType.Ack
    assert msg.sequence_number == 0
    assert msg.target_participant_id == 2
    assert msg.source_participant_id == 3
    assert msg.channel_id == 0xFFFF
    assert msg.flags.need_ack is False
    assert msg.flags.is_fragment is False

    assert len(packed) == 26
    assert packed == unhexlify(
        b'd00d00000000000000000002000000038001000000000000ffff'
    )


def test_power_on(packets, crypto):
    bin_name = 'poweron_request'
    msg = factory.power_on(liveid='FD00112233FFEE66')
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.PowerOnRequest
    assert msg.header.version == 0
    assert msg.unprotected_payload.liveid == 'FD00112233FFEE66'

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]


def test_discovery_request(packets, crypto):
    bin_name = 'discovery_request'
    msg = factory.discovery(client_type=enum.ClientType.Android)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.DiscoveryRequest
    assert msg.header.version == 0

    assert msg.unprotected_payload.flags == 0
    assert msg.unprotected_payload.client_type == enum.ClientType.Android
    assert msg.unprotected_payload.minimum_version == 0
    assert msg.unprotected_payload.maximum_version == 2

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]


def test_connect_request(packets, crypto):
    bin_name = 'connect_request'
    msg = factory.connect(
        sg_uuid=uuid.UUID('{de305d54-75b4-431b-adb2-eb6b9e546014}'),
        public_key_type=enum.PublicKeyType.EC_DH_P256,
        public_key=b'\xFF' * 64,
        iv=unhexlify('2979d25ea03d97f58f46930a288bf5d2'),
        userhash='deadbeefdeadbeefde',
        jwt='dummy_token',
        msg_num=0,
        num_start=0,
        num_end=2
    )
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.ConnectRequest

    assert str(msg.unprotected_payload.sg_uuid) == 'de305d54-75b4-431b-adb2-eb6b9e546014'
    assert msg.unprotected_payload.public_key_type == enum.PublicKeyType.EC_DH_P256
    assert msg.unprotected_payload.public_key == b'\xFF' * 64
    assert msg.unprotected_payload.iv == unhexlify(
        '2979d25ea03d97f58f46930a288bf5d2'
    )

    assert msg.protected_payload.userhash == 'deadbeefdeadbeefde'
    assert msg.protected_payload.jwt == 'dummy_token'
    assert msg.protected_payload.connect_request_num == 0
    assert msg.protected_payload.connect_request_group_start == 0
    assert msg.protected_payload.connect_request_group_end == 2

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]


def test_message_fragment(packets, crypto):
    data = unhexlify(
        b'6a5297cd00424d6963726f736f66742e426c75726179506c617965725f3877656b79623364386262'
        b'77652158626f782e426c75726179506c617965722e4170706c69636174696f6e0008883438303036'
        b'31303036433030364630303230303035343030363830303635303032303030343630303631303036'
        b'43303036433030323030303646303036363030323030303532303036353030363130303633303036'
        b'38303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'30303030303030303030303030303030303030303030303030303030303030303030303030303030'
        b'303030303030303030303030303030303030303030303030'
    )

    msg = factory.message_fragment(
        msg_type=enum.MessageType.MediaState,
        sequence_begin=22,
        sequence_end=25,
        data=data,
        need_ack=True,
        source_participant_id=0,
        target_participant_id=31,
        channel_id=148
    )

    msg.header(sequence_number=22)

    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.MediaState

    assert msg.protected_payload.sequence_begin == 22
    assert msg.protected_payload.sequence_end == 25
    assert msg.protected_payload.data == data

    assert isinstance(packed, bytes) is True
    assert len(packed) == 1098

    assert packed == packets['fragment_media_state_0']


def test_acknowledge(packets, crypto):
    bin_name = 'acknowledge'
    msg = factory.acknowledge(
        low_watermark=0,
        processed_list=[1],
        rejected_list=[]
    )
    msg.header(
        sequence_number=1,
        target_participant_id=31,
        channel_id=1152921504606846976
    )
    msg.header.flags(version=2, need_ack=False)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.Ack

    assert msg.protected_payload.low_watermark == 0
    assert msg.protected_payload.processed_list == [1]
    assert msg.protected_payload.rejected_list == []

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]


def test_json(packets, crypto):
    bin_name = 'json'
    msg = factory.json(
        text={'request': 'GetConfiguration', 'msgid': '2ed6c0fd.2'}
    )
    msg.header(
        sequence_number=11,
        source_participant_id=31,
        channel_id=151
    )
    msg.header.flags(need_ack=True)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.Json

    assert msg.protected_payload.text == {
        'request': 'GetConfiguration', 'msgid': '2ed6c0fd.2'
    }

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]


def test_disconnect(packets, crypto):
    bin_name = 'disconnect'
    msg = factory.disconnect(
        reason=enum.DisconnectReason.Unspecified,
        error_code=0
    )
    msg.header(
        sequence_number=57,
        source_participant_id=31
    )
    msg.header.flags(need_ack=False)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.Disconnect

    assert msg.protected_payload.reason == enum.DisconnectReason.Unspecified
    assert msg.protected_payload.error_code == 0

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]


def test_power_off(packets, crypto):
    bin_name = 'power_off'
    msg = factory.power_off(liveid='FD00112233FFEE66')
    msg.header(
        sequence_number=1882,
        source_participant_id=2
    )
    msg.header.flags(need_ack=True)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.PowerOff

    assert msg.protected_payload.liveid == 'FD00112233FFEE66'

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]


def test_local_join(packets, crypto):
    class TestClientInfo(object):
        DeviceType = enum.ClientType.Android
        NativeWidth = 600
        NativeHeight = 1024
        DpiX = 160
        DpiY = 160
        DeviceCapabilities = constants.DeviceCapabilities.All
        ClientVersion = 133713371
        OSMajor = 42
        OSMinor = 0
        DisplayName = 'package.name.here'

    bin_name = 'local_join'
    msg = factory.local_join(client_info=TestClientInfo)
    msg.header(
        sequence_number=1,
        source_participant_id=31
    )
    msg.header.flags(version=0)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.LocalJoin

    assert msg.protected_payload.device_type == enum.ClientType.Android
    assert msg.protected_payload.native_width == 600
    assert msg.protected_payload.native_height == 1024
    assert msg.protected_payload.dpi_x == 160
    assert msg.protected_payload.dpi_y == 160
    assert msg.protected_payload.device_capabilities == constants.DeviceCapabilities.All
    assert msg.protected_payload.client_version == 133713371
    assert msg.protected_payload.os_major_version == 42
    assert msg.protected_payload.os_minor_version == 0
    assert msg.protected_payload.display_name == 'package.name.here'

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]


def test_title_launch(packets, crypto):
    bin_name = 'title_launch'
    msg = factory.title_launch(
        location=enum.ActiveTitleLocation.Fill,
        uri='ms-xbl-0D174C79://default/'
    )
    msg.header(
        sequence_number=685,
        source_participant_id=32
    )
    msg.header.flags(need_ack=True)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.TitleLaunch

    assert msg.protected_payload.location == enum.ActiveTitleLocation.Fill
    assert msg.protected_payload.uri == 'ms-xbl-0D174C79://default/'

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]


def test_start_channel(packets, crypto):
    bin_name = 'start_channel_request'
    msg = factory.start_channel(
        channel_request_id=1,
        title_id=0,
        service=constants.MessageTarget.SystemInputUUID,
        activity_id=0
    )
    msg.header(
        sequence_number=2,
        source_participant_id=31
    )
    msg.header.flags(need_ack=True)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.StartChannelRequest

    assert msg.protected_payload.channel_request_id == 1
    assert msg.protected_payload.title_id == 0
    assert msg.protected_payload.service == constants.MessageTarget.SystemInputUUID
    assert msg.protected_payload.activity_id == 0

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]


def test_stop_channel(packets, crypto):
    msg = factory.stop_channel(channel_id=2)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.StopChannel

    assert msg.protected_payload.target_channel_id == 2

    assert len(packed) == 74
    assert packed == unhexlify(
        b'd00d0008000000000000000000000000802800000000000000001b0783367a4426'
        b'c2e0775916685c072df0380ee320925842716d595ced4b68f8a9bad01a5301be7e'
        b'd7d3b4e25a03b728'
    )


def test_gamepad(packets, crypto):
    bin_name = 'gamepad'
    msg = factory.gamepad(
        timestamp=0,
        buttons=enum.GamePadButton.PadB,
        l_trigger=0.0,
        r_trigger=0.0,
        l_thumb_x=0.0,
        r_thumb_x=0.0,
        l_thumb_y=0.0,
        r_thumb_y=0.0
    )

    msg.header(
        sequence_number=79,
        source_participant_id=41,
        channel_id=180
    )
    msg.header.flags(need_ack=False)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.Gamepad

    assert msg.protected_payload.timestamp == 0
    assert msg.protected_payload.buttons == enum.GamePadButton.PadB
    assert msg.protected_payload.left_trigger == 0.0
    assert msg.protected_payload.right_trigger == 0.0
    assert msg.protected_payload.left_thumbstick_x == 0.0
    assert msg.protected_payload.right_thumbstick_x == 0.0
    assert msg.protected_payload.left_thumbstick_y == 0.0
    assert msg.protected_payload.right_thumbstick_y == 0.0

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]


def test_unsnap(packets, crypto):
    msg = factory.unsnap(unknown=1)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.Unsnap

    assert msg.protected_payload.unk == 1

    assert len(packed) == 74
    assert packed == unhexlify(
        b'd00d000100000000000000000000000080370000000000000000738d3b76a69a34'
        b'bbf732755035fe77672e5ea1432f3e278189d7756f62254ebe790dfe084ba43067'
        b'bf3f4b2546c1f882'
    )


def test_gamedvr_record(packets, crypto):
    bin_name = 'gamedvr_record'
    msg = factory.game_dvr_record(
        start_time_delta=-60,
        end_time_delta=0
    )
    msg.header(
        sequence_number=70,
        source_participant_id=1
    )
    msg.header.flags(need_ack=True)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.GameDvrRecord

    assert msg.protected_payload.start_time_delta == -60
    assert msg.protected_payload.end_time_delta == 0

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]


def test_media_command(packets, crypto):
    bin_name = 'media_command'
    msg = factory.media_command(
        request_id=0,
        title_id=274278798,
        command=enum.MediaControlCommand.FastForward,
        seek_position=0
    )
    msg.header(
        sequence_number=597,
        source_participant_id=32,
        channel_id=153
    )
    msg.header.flags(need_ack=True)
    packed = _pack(msg, crypto)

    assert msg.header.pkt_type == enum.PacketType.Message
    assert msg.header.flags.msg_type == enum.MessageType.MediaCommand

    assert msg.protected_payload.request_id == 0
    assert msg.protected_payload.title_id == 274278798
    assert msg.protected_payload.command == enum.MediaControlCommand.FastForward

    assert len(packed) == len(packets[bin_name])
    assert packed == packets[bin_name]



"""
Cryptography portion used for sending Smartglass message

Depending on the foreign public key type, the following Elliptic curves
can be used:

* Prime 256R1
* Prime 384R1
* Prime 521R1

1. On Discovery, the console responds with a DiscoveryResponse
   including a certificate, this certificate holds the console's
   `public key`.
2. The Client generates appropriate `elliptic curve` and derives the
   `shared secret` using `console's public key`
3. The `shared secret` is salted via 2x hashes, see `kdf_salts`
4. The `salted shared secret` is hashed using `SHA-512`
5. The `salted & hashed shared secret` is split into the following
   individual keys:

    * bytes  0-16: Encryption key (AES 128-CBC)
    * bytes 16-32: Initialization Vector key
    * bytes 32-64: Hashing key (HMAC SHA-256)

6. The resulting `public key` from this :class:`Crypto` context is
   sent with the ConnectRequest message to the console
"""
import os
import hmac
import hashlib
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from xbox.sg.enum import PublicKeyType
from binascii import unhexlify


CURVE_MAP = {
    PublicKeyType.EC_DH_P256: ec.SECP256R1,
    PublicKeyType.EC_DH_P384: ec.SECP384R1,
    PublicKeyType.EC_DH_P521: ec.SECP521R1
}


PUBLIC_KEY_TYPE_MAP = {v: k for k, v in CURVE_MAP.items()}


class SaltType(object):
    """
    Define whether Salt is pre- or appended
    """
    Prepend = 1
    Append = 2


class Salt(object):
    def __init__(self, value, salt_type=SaltType.Prepend):
        """
        Handle salting of ECDH shared secret

        Args:
            value (bytes): Salting bytes
            salt_type (:obj:`SaltType`): Salt Type
        """
        self.value = value
        self.type = salt_type

    def apply(self, data):
        """
        Appends or prepends salt bytes to data

        Args:
            data (bytes): Data to be salted

        Returns:
            bytes: Salted data
        """
        if self.type == SaltType.Prepend:
            return self.value + data
        if self.type == SaltType.Append:
            return data + self.value
        raise ValueError("Unknown salt type: " + str(self.type))


class Crypto(object):
    _backend = default_backend()

    _kdf_salts = (
        Salt(unhexlify('D637F1AAE2F0418C'), SaltType.Prepend),
        Salt(unhexlify('A8F81A574E228AB7'), SaltType.Append)
    )

    def __init__(self, foreign_public_key, privkey=None, pubkey=None):
        """
        Initialize Crypto context via the foreign public key of the console.
        The public key is part of the console certificate.

        Args:
            foreign_public_key (:obj:`ec.EllipticCurvePublicKey`):
                The console's public key
            privkey (:obj:`ec.EllipticCurvePrivateKey`): Optional private key
            pubkey (:obj:`ec.EllipticCurvePublicKey`): Optional public key
        """

        if not isinstance(foreign_public_key, ec.EllipticCurvePublicKey):
            raise ValueError("Unsupported public key format, \
                expected EllipticCurvePublicKey")

        if privkey and not isinstance(privkey, ec.EllipticCurvePrivateKey):
            raise ValueError("Unsupported private key format, \
                expected EllipticCurvePrivateKey")
        elif not privkey:
            privkey = ec.generate_private_key(
                foreign_public_key.curve, self._backend
            )

        if pubkey and not isinstance(pubkey, ec.EllipticCurvePublicKey):
            raise ValueError("Unsupported public key format, \
                expected EllipticCurvePublicKey")
        elif not pubkey:
            self.pubkey = privkey.public_key()
        else:
            self.pubkey = pubkey

        secret = privkey.exchange(ec.ECDH(), foreign_public_key)
        salted_secret = secret
        for salt in self._kdf_salts:
            salted_secret = salt.apply(salted_secret)

        self._expanded_secret = hashlib.sha512(salted_secret).digest()
        self._encrypt_key = self._expanded_secret[:16]
        self._iv_key = self._expanded_secret[16:32]
        self._hash_key = self._expanded_secret[32:]

        self._pubkey_type = PUBLIC_KEY_TYPE_MAP[type(self.pubkey.curve)]
        self._pubkey_bytes = self.pubkey.public_bytes(
            format=PublicFormat.UncompressedPoint,
            encoding=Encoding.X962)[1:]
        self._foreign_pubkey = foreign_public_key

    @property
    def shared_secret(self):
        """
        Shared secret

        Returns:
            bytes: Shared secret
        """
        return self._expanded_secret

    @property
    def pubkey_type(self):
        """
        Public Key Type aka. keystrength

        Returns:
            :obj:`.PublicKeyType`: Public Key Type

        """
        return self._pubkey_type

    @property
    def pubkey_bytes(self):
        """
        Public Key Bytes (minus the first identifier byte!)

        Returns:
            bytes: Public key
        """
        return self._pubkey_bytes

    @property
    def foreign_pubkey(self):
        """
        Foreign key that was used to generate this crypto context

        Returns:
            :obj:`ec.EllipticCurvePublicKey`: Console's public key
        """
        return self._foreign_pubkey

    @classmethod
    def from_bytes(cls, foreign_public_key, public_key_type=None):
        """
        Initialize Crypto context with foreign public key in
        bytes / hexstring format.

        Args:
            foreign_public_key (bytes): Console's public key
            public_key_type (:obj:`.PublicKeyType`): Public Key Type

        Returns:
            :obj:`.Crypto`: Instance
        """

        if not isinstance(foreign_public_key, bytes):
            raise ValueError("Unsupported foreign public key format, \
                expected bytes")

        if public_key_type is None:
            keylen = len(foreign_public_key)
            if keylen == 0x41:
                public_key_type = PublicKeyType.EC_DH_P256
            elif keylen == 0x61:
                public_key_type = PublicKeyType.EC_DH_P384
            elif keylen == 0x85:
                public_key_type = PublicKeyType.EC_DH_P521
            else:
                raise ValueError("Invalid public keylength")

        curve = CURVE_MAP[public_key_type]
        foreign_public_key = ec.EllipticCurvePublicKey.from_encoded_point(
            curve(), foreign_public_key
        )
        return cls(foreign_public_key)

    @classmethod
    def from_shared_secret(cls, shared_secret):
        """
        Set up crypto context with shared secret

        Args:
            shared_secret (bytes): The shared secret

        Returns:
            :obj:`.Crypto`: Instance
        """
        if not isinstance(shared_secret, bytes):
            raise ValueError("Unsupported shared secret format, \
                expected bytes")
        elif len(shared_secret) != 64:
            raise ValueError("Unsupported shared secret length, \
                expected 64 bytes")

        # We need a dummy foreign key
        dummy_key = unhexlify(
            '041db1e7943878b28c773228ebdcfb05b985be4a386a55f50066231360785f61b'
            '60038caf182d712d86c8a28a0e7e2733a0391b1169ef2905e4e21555b432b262d'
        )
        ctx = cls.from_bytes(dummy_key)
        # Overwriting shared secret
        ctx._encrypt_key = shared_secret[:16]
        ctx._iv_key = shared_secret[16:32]
        ctx._hash_key = shared_secret[32:]
        return ctx

    def generate_iv(self, seed=None):
        """
        Generates an IV to be used in encryption/decryption

        Args:
            seed (bytes): An optional IV seed

        Returns:
            bytes: Initialization Vector
        """
        if seed:
            return self._encrypt(key=self._iv_key, iv=None, data=seed)
        return os.urandom(16)

    def encrypt(self, iv, plaintext):
        """
        Encrypts plaintext with AES-128-CBC

        No padding is added here, data has to be aligned to
        block size (16 bytes).

        Args:
            iv (bytes): The IV to use. None where no IV is used.
            plaintext (bytes): The plaintext to encrypt.

        Returns:
            bytes: Encrypted Data
        """
        return Crypto._encrypt(self._encrypt_key, iv, plaintext)

    def decrypt(self, iv, ciphertext):
        """
        Decrypts ciphertext

        No padding is removed here.

        Args:
            iv (bytes): The IV to use. None where no IV is used.
            ciphertext (bytes): The hex representation of a ciphertext
                                to be decrypted

        Returns:
            bytes: Decrypted data
        """
        return Crypto._decrypt(self._encrypt_key, iv, ciphertext)

    def hash(self, data):
        """
        Securely hashes data with HMAC SHA-256

        Args:
            data (bytes): The data to securely hash.

        Returns:
            bytes: Hashed data
        """
        return Crypto._secure_hash(self._hash_key, data)

    def verify(self, data, secure_hash):
        """
        Verifies that the given data generates the given secure_hash

        Args:
            data (bytes): The data to validate.
            secure_hash (bytes): The secure hash to validate against.

        Returns:
            bool: True on success, False otherwise
        """
        return secure_hash == self.hash(data)

    @staticmethod
    def _secure_hash(key, data):
        return hmac.new(key, data, hashlib.sha256).digest()

    @staticmethod
    def _encrypt(key, iv, data):
        return Crypto._crypt(key=key, iv=iv, encrypt=True, data=data)

    @staticmethod
    def _decrypt(key, iv, data):
        return Crypto._crypt(key=key, iv=iv, encrypt=False, data=data)

    @staticmethod
    def _crypt(key, iv, encrypt, data):
        if not iv:
            iv = b'\x00' * 16
        cipher = Cipher(
            algorithms.AES(key), modes.CBC(iv), backend=Crypto._backend
        )
        if encrypt:
            cryptor = cipher.encryptor()
        else:
            cryptor = cipher.decryptor()

        return cryptor.update(data) + cryptor.finalize()


class Padding(object):
    """
    Padding base class.
    """
    @staticmethod
    def size(length, alignment):
        """
        Calculate needed padding size.

        Args:
            length (int): Data size
            alignment (int): Data alignment

        Returns:
            int: Padding size
        """
        overlap = length % alignment
        if overlap:
            return alignment - overlap
        else:
            return 0

    @staticmethod
    def pad(payload, alignment):
        """
        Abstract method to override

        Args:
            payload (bytes): Data blob
            alignment (int): Data alignment

        Returns:
            bytes: Data with padding bytes
        """
        raise NotImplementedError()

    @staticmethod
    def remove(payload):
        """
        Common method for removing padding from data blob.

        Args:
            payload (bytes): Padded data.

        Returns:
            bytes: Data with padding bytes removed
        """
        pad_count = payload[-1]
        return payload[:-pad_count]


class PKCS7Padding(Padding):
    @staticmethod
    def pad(payload, alignment):
        """
        Add PKCS#7 padding to data blob.

        Args:
            payload (bytes): Data blob
            alignment (int): Data alignment

        Returns:
            bytes: Data with padding bytes
        """
        size = Padding.size(len(payload), alignment)
        if size == 0:
            return payload
        else:
            return payload + (size * chr(size).encode())


class ANSIX923Padding(Padding):
    @staticmethod
    def pad(payload, alignment):
        """
        Add ANSI.X923 padding to data blob.

        Args:
            payload (bytes): Data blob
            alignment (int): Data alignment

        Returns:
            bytes: Data with padding bytes
        """
        size = Padding.size(len(payload), alignment)
        if size == 0:
            return payload
        else:
            return payload + ((size - 1) * b'\x00') + chr(size).encode()
